# yapf: disable
import logging
import numpy as np
import prettytable
import torch
from mmcv.runner import build_optimizer
from mmcv.runner.hooks import Hook
from typing import List, Union
from xrprimer.utils.log_utils import get_logger

from xrmocap.core.hook.smplify_hook.builder import (
    SMPLifyBaseHook, build_smplify_hook,
)
from xrmocap.model.body_model.builder import build_body_model
from xrmocap.transform.convention.keypoints_convention import (  # noqa:E501
    get_keypoint_idx, get_keypoint_idxs_by_part,
)
from .handler.base_handler import BaseHandler, BaseInput
from .handler.builder import build_handler
from .optimizable_parameters import OptimizableParameters

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
# yapf: enable


class SMPLify(object):
    """Re-implementation of SMPLify with extended features."""
    OPTIM_PARAM = ['global_orient', 'transl', 'body_pose', 'betas']

    def __init__(self,
                 body_model: Union[dict, torch.nn.Module],
                 stages: dict,
                 optimizer: dict,
                 handlers: List[Union[dict, BaseHandler]],
                 n_epochs: int = 1,
                 use_one_betas_per_video: bool = False,
                 ignore_keypoints: List[str] = None,
                 device: Union[torch.device, str] = 'cuda',
                 hooks: List[Union[dict, SMPLifyBaseHook]] = [],
                 verbose: bool = False,
                 info_level: Literal['stage', 'step'] = 'step',
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Re-implementation of SMPLify with extended features.

        Args:
            body_model (Union[dict, torch.nn.Module]):
                An instance of SMPL body_model or a config dict
                of SMPL body_model.
            stages (dict):
                A config dict of registration stages.
            optimizer (dict):
                A config dict of optimizer.
            handlers (List[Union[dict, BaseHandler]]):
                A list of handlers, each element is an instance of
                subclass of BaseHandler, or a config dict of handler.
            n_epochs (int, optional):
                Number of epochs. Defaults to 1.
            use_one_betas_per_video (bool, optional):
                Whether to use the same beta parameters
                for all frames in a single video sequence.
                Defaults to False.
            ignore_keypoints (List[str], optional):
                A list of keypoint names to ignore in keypoint
                loss computation. Defaults to None.
            device (Union[torch.device, str], optional):
                Device in pytorch. Defaults to 'cuda'.
            hooks (List[Union[dict, SMPLifyBaseHook]], optional):
                A list of hooks, each element is an instance of
                subclass of SMPLifyBaseHook, or a config dict of hook.
                Defaults to an empty list.
            verbose (bool, optional):
                Whether to print individual losses during registration.
                Defaults to False.
            info_level (Literal['stage', 'step']):
                Whether to print information every stage or every step.
                Defaults to stage.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.

        Raises:
            TypeError: type of body_model should be
            either dict or torch.nn.Module.
        """
        self.logger = get_logger(logger)
        self.info_level = info_level
        self.use_one_betas_per_video = use_one_betas_per_video
        self.n_epochs = n_epochs
        self.device = device
        self.stage_config = stages
        self.optimizer = optimizer
        self.hooks = []

        # initialize body model
        if isinstance(body_model, dict):
            self.body_model = build_body_model(body_model).to(self.device)
        elif isinstance(body_model, torch.nn.Module):
            self.body_model = body_model.to(self.device)
        else:
            self.logger.error('body_model should be either dict or ' +
                              f'torch.nn.Module, but got {type(body_model)}.')
            raise TypeError

        self.ignore_keypoints = ignore_keypoints \
            if ignore_keypoints is not None \
            else []
        self.verbose = verbose
        self.loss_handlers = []
        # init handlers
        for handler in handlers:
            if isinstance(handler, dict):
                handler['device'] = self.device
                handler = build_handler(handler)
            self.loss_handlers.append(handler)
        # init hooks
        for hook in hooks:
            if isinstance(hook, dict):
                hook = build_smplify_hook(hook)
            self.register_hook(hook)
        self.__set_keypoint_indexes__()
        self.__stage_kwargs_warned__ = False

    def register_hook(self, hook: SMPLifyBaseHook):
        """Register a hook into the hook list.

        The hook will be inserted into a priority queue.

        Args:
            hook (SMPLifyBaseHook):
                The hook to be registered.
        """
        assert isinstance(hook, Hook)
        hook.priority = 'NORMAL'
        self.hooks.append(hook)

    def call_hook(self, fn_name: str, **kwargs):
        """Call all hooks.

        Args:
            fn_name (str): The function name in each hook to be called, such as
                "before_optimize".

        kwargs:
            Keyword args required by the hook.
        """
        for hook in self.hooks:
            getattr(hook, fn_name)(self, **kwargs)

    def __prepare_optimizable_parameters__(self, init_dict: dict,
                                           batch_size: int) -> dict:
        """Prepare optimizable parameters in batch for registrant. If some of
        the parameters can be found in init_dict, use them for initialization.

        Args:
            init_dict (dict):
                A dict of init parameters. init_dict.keys() is a
                sub-set of self.__class__.OPTIM_PARAM.
            batch_size (int)

        Returns:
            dict:
                A dict of optimizable parameters, whose keys are
                self.__class__.OPTIM_PARAM and values are
                Tensors in batch.
        """
        ret_dict = {}
        for key in self.__class__.OPTIM_PARAM:
            if key in init_dict:
                init_param = init_dict[key]
            else:
                init_param = None
            ret_param = self.__match_init_batch_size__(
                init_param=init_param,
                default_param=getattr(self.body_model, key),
                batch_size=batch_size)
            ret_dict[key] = ret_param
        if self.use_one_betas_per_video and 'betas' not in init_dict:
            betas = torch.zeros(1, self.body_model.betas.shape[-1]).to(
                self.device)
            ret_dict['betas'] = betas
        return ret_dict

    def __call__(self,
                 input_list: List[BaseInput],
                 init_param_dict: dict = {},
                 return_verts: bool = False,
                 return_joints: bool = False,
                 return_full_pose: bool = False,
                 return_losses: bool = False) -> dict:
        """Run registration.

        Args:
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            init_param_dict (dict, optional):
                A dict of init parameters. init_dict.keys() is a
                sub-set of self.__class__.OPTIM_PARAM.
                Defaults to an empty dict.
            return_verts (bool, optional):
                Whether to return vertices. Defaults to False.
            return_joints (bool, optional):
                Whether to return joints. Defaults to False.
            return_full_pose (bool, optional):
                Whether to return full pose. Defaults to False.
            return_losses (bool, optional):
                Whether to return loss dict. Defaults to False.

        Notes:
            B: batch size
            K: number of keypoints
            D: shape dimension

        Returns:
            dict:
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
        """
        batch_size = None
        for input_instance in input_list:
            tmp_batch_size = input_instance.get_batch_size()
            if batch_size is None:
                batch_size = tmp_batch_size
            elif batch_size != tmp_batch_size:
                self.logger.error('Batch size varies among input_list.\n' +
                                  f'Target batch size: {batch_size}\n' +
                                  f'{input_instance.handler_key} batch size:' +
                                  f' {tmp_batch_size}')
                raise ValueError

        optim_param = self.__prepare_optimizable_parameters__(
            init_param_dict, batch_size)

        hook_kwargs = dict(input_list=input_list, optim_param=optim_param)
        self.call_hook('before_optimize', **hook_kwargs)

        for i in range(self.n_epochs):
            for stage_idx, stage_config in enumerate(self.stage_config):
                self.__optimize_stage__(
                    input_list=input_list,
                    optim_param=optim_param,
                    epoch_idx=i,
                    stage_idx=stage_idx,
                    **stage_config,
                )

        hook_kwargs = dict(input_list=input_list, optim_param=optim_param)
        self.call_hook('after_optimize', **hook_kwargs)
        # collate results
        ret = optim_param

        if return_verts or return_joints or \
                return_full_pose or return_losses:
            eval_ret = self.evaluate(
                input_list=input_list,
                optim_param=optim_param,
                return_verts=return_verts,
                return_full_pose=return_full_pose,
                return_joints=return_joints,
                reduction_override='none'  # sample-wise loss
            )

            if return_verts:
                ret['vertices'] = eval_ret['vertices']
            if return_joints:
                ret['joints'] = eval_ret['joints']
            if return_full_pose:
                ret['full_pose'] = eval_ret['full_pose']
            if return_losses:
                for k in eval_ret.keys():
                    if 'loss' in k:
                        ret[k] = eval_ret[k]

        for k, v in ret.items():
            if isinstance(v, torch.Tensor):
                ret[k] = v.detach().clone()
        return ret

    def __optimize_stage__(self,
                           input_list: List[BaseInput],
                           optim_param: dict,
                           epoch_idx: int = -1,
                           stage_idx: int = -1,
                           use_shoulder_hip_only: bool = False,
                           body_weight: float = 1.0,
                           n_iter: int = 1,
                           ftol: float = 1e-4,
                           **kwargs) -> None:
        """Optimize a stage of body model parameters according to
        configuration.

        Args:
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            param_dict (dict):
                A dict of optimizable parameters, whose keys are
                self.__class__.OPTIM_PARAM and values are
                Tensors in batch.
            epoch_idx (int, optional):
                The index of this epoch. For hook and verbose only,
                it will not influence the optimize result.
                Defaults to -1.
            stage_idx (int, optional):
                The index of this stage. For hook and verbose only,
                it will not influence the optimize result.
                Defaults to -1.
            use_shoulder_hip_only (bool, optional):
                Keypoints weight argument.
                Whether to use only shoulder and hip
                keypoints for loss computation. This is useful in the
                warming-up stage to find a reasonably good initialization.
                Defaults to False.
            body_weight (float, optional):
                Keypoints weight argument.
                Weight of body keypoints. Body part segmentation
                definition is included in the HumanData convention.
                Defaults to 1.0.
            n_iter (int, optional):
                Number of iterations. Defaults to 1.
            ftol (float, optional):
                Defaults to 1e-4.

        kwargs:
            Parameter fit flag and stage control keyword arguments.
            Parameter fit args includes fit_{param_key}, while param_key
            is in self.__class__.OPTIM_PARAM. If fit_{param_key} not found
            in kwargs, use default value True.
            Stage control args includes override weights
            of each loss handler.

        Notes:
            B: batch size
            K: number of keypoints
            D: shape dimension
        """
        stage_config = dict(
            epoch_idx=epoch_idx,
            stage_idx=stage_idx,
            use_shoulder_hip_only=use_shoulder_hip_only,
            body_weight=body_weight,
            n_iter=n_iter,
            ftol=ftol)
        stage_config.update(kwargs)
        hook_kwargs = dict(
            input_list=input_list,
            optim_param=optim_param,
            stage_config=stage_config)
        self.call_hook('before_stage', **hook_kwargs)

        kwargs = kwargs.copy()
        parameters = OptimizableParameters()
        for key, value in optim_param.items():
            fit_flag = kwargs.pop(f'fit_{key}', True)
            parameters.add_param(key=key, param=value, fit_param=fit_flag)
        optimizer = build_optimizer(parameters, self.optimizer)

        pre_loss = None
        for iter_idx in range(n_iter):

            def closure():
                optimizer.zero_grad()
                betas_video = self.__expand_betas__(
                    batch_size=optim_param['body_pose'].shape[0],
                    betas=optim_param['betas'])
                expanded_param = {}
                expanded_param.update(optim_param)
                expanded_param['betas'] = betas_video
                loss_dict = self.evaluate(
                    input_list=input_list,
                    optim_param=expanded_param,
                    use_shoulder_hip_only=use_shoulder_hip_only,
                    body_weight=body_weight,
                    **kwargs)

                loss = loss_dict['total_loss']
                loss.backward()
                return loss

            loss = optimizer.step(closure)
            if iter_idx > 0 and pre_loss is not None and ftol > 0:
                loss_rel_change = self.__compute_relative_change__(
                    pre_loss, loss.item())
                if loss_rel_change < ftol:
                    if self.verbose:
                        self.logger.info(
                            f'[ftol={ftol}] Early stop at {iter_idx} iter!')
                    break
            pre_loss = loss.item()

        stage_config = dict(
            use_shoulder_hip_only=use_shoulder_hip_only,
            body_weight=body_weight,
            n_iter=n_iter,
            ftol=ftol)
        stage_config.update(kwargs)

        betas_video = self.__expand_betas__(
            batch_size=optim_param['body_pose'].shape[0],
            betas=optim_param['betas'])
        expanded_param = optim_param.copy()
        expanded_param['betas'] = betas_video
        loss_dict = self.evaluate(
            input_list=input_list,
            optim_param=expanded_param,
            use_shoulder_hip_only=use_shoulder_hip_only,
            body_weight=body_weight,
            **kwargs)
        hook_kwargs = dict(
            input_list=input_list,
            optim_param=optim_param,
            stage_config=stage_config,
            loss_dict=loss_dict)
        self.call_hook('after_stage', **hook_kwargs)

    def evaluate(self,
                 input_list: List[BaseInput],
                 optim_param: dict,
                 return_verts: bool = False,
                 return_full_pose: bool = False,
                 return_joints: bool = False,
                 use_shoulder_hip_only: bool = False,
                 body_weight: float = 1.0,
                 reduction_override: Literal['mean', 'sum', 'none'] = None,
                 **kwargs) -> dict:
        """Evaluate fitted parameters through loss computation. This function
        serves two purposes: 1) internally, for loss backpropagation 2)
        externally, for fitting quality evaluation.

        Args:
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            optim_param (dict):
                A dict of optimizable parameters, whose keys are
                self.__class__.OPTIM_PARAM and values are
                Tensors in batch.
            return_verts (bool, optional):
                Whether to return vertices. Defaults to False.
            return_joints (bool, optional):
                Whether to return joints. Defaults to False.
            return_full_pose (bool, optional):
                Whether to return full pose. Defaults to False.
            return_losses (bool, optional):
                Whether to return loss dict. Defaults to False.
            use_shoulder_hip_only (bool, optional):
                Keypoints weight argument.
                Whether to use only shoulder and hip
                keypoints for loss computation. This is useful in the
                warming-up stage to find a reasonably good initialization.
                Defaults to False.
            body_weight (float, optional):
                Keypoints weight argument.
                Weight of body keypoints. Body part segmentation
                definition is included in the HumanData convention.
                Defaults to 1.0.
            reduction_override (Literal['mean', 'sum', 'none'], optional):
                The reduction method. If given, it will
                override the original reduction method of the loss.
                Defaults to None.

        kwargs:
            Stage control keyword arguments, including override weights
            of eache loss handler.

        Notes:
            B: batch size
            K: number of keypoints
            D: shape dimension

        Returns:
            dict:
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
        """
        hook_kwargs = dict(input_list=input_list, optim_param=optim_param)
        self.call_hook('before_evaluate', **hook_kwargs)
        ret = {}

        require_verts = self.__check_verts_requirement__(
            input_list, **kwargs) or return_verts
        body_model_kwargs = dict(
            return_verts=require_verts, return_full_pose=return_full_pose)
        body_model_kwargs.update(optim_param)
        body_model_output = self.body_model(**body_model_kwargs)

        model_joints = body_model_output['joints']
        model_joints_convention = self.body_model.keypoint_convention
        model_joints_weights = self.get_keypoint_weight(
            use_shoulder_hip_only=use_shoulder_hip_only,
            body_weight=body_weight)
        model_vertices = body_model_output.get('vertices', None)

        loss_dict = self.__compute_loss__(
            input_list=input_list,
            model_joints=model_joints,
            model_joints_convention=model_joints_convention,
            model_joints_weights=model_joints_weights,
            model_vertices=model_vertices,
            reduction_override=reduction_override,
            optim_param=optim_param,
            **kwargs)
        ret.update(loss_dict)

        if return_verts:
            ret['vertices'] = body_model_output['vertices']
        if return_full_pose:
            ret['full_pose'] = body_model_output['full_pose']
        if return_joints:
            ret['joints'] = model_joints

        hook_kwargs = dict(
            input_list=input_list,
            optim_param=optim_param,
            loss_dict=loss_dict)
        self.call_hook('after_evaluate', **hook_kwargs)

        return ret

    def __compute_loss__(self,
                         input_list: List[BaseInput],
                         model_joints: torch.Tensor = None,
                         model_joints_convention: str = '',
                         model_joints_weights: torch.Tensor = None,
                         model_vertices: torch.Tensor = None,
                         optim_param: dict = None,
                         reduction_override: Literal['mean', 'sum',
                                                     'none'] = None,
                         **kwargs) -> dict:
        """Loss computation.

        Args:
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            model_joints (torch.Tensor):
                Output joints from self.body_model, of shape (B, K, 3).
                Defaults to None.
            model_joints_convention (str):
                Convention name of model_joints. Defaults to ''.
            model_joints_weights (torch.Tensor):
                If given, each joint has its own weight, in shape (K, ).
                Defaults to None.
            model_vertices (torch.Tensor):
                Output joints from self.body_model. Defaults to None.
            param_dict (dict):
                A dict of optimizable parameters, whose keys are
                self.__class__.OPTIM_PARAM and values are
                Tensors in batch.
            reduction_override (Literal['mean', 'sum', 'none'], optional):
                The reduction method. If given, it will
                override the original reduction method of the loss.
                Defaults to None.

        kwargs:
            Stage control keyword arguments, including override weights
            of each loss handler.

        Notes:
            B: batch size
            K: number of keypoints
            D: shape dimension

        Returns:
            dict: A dict that contains all losses.
        """
        if model_joints_weights is None and \
                model_joints is not None:
            model_joints_weights = torch.ones_like(model_joints[0, :, 0])
        init_handler_input = {
            'model_vertices': model_vertices,
            'model_joints': model_joints,
            'model_joints_convention': model_joints_convention,
            'model_joints_weights': model_joints_weights,
        }
        init_handler_input.update(optim_param)
        losses = {}
        # backup kwargs for pop
        kwargs = kwargs.copy()
        for handler in self.loss_handlers:
            handler_key = handler.handler_key
            # Get args from stage controller.
            # If used, pop it.
            loss_weight_override = kwargs.pop(f'{handler_key}_weight', None)
            reduction_override = reduction_override \
                if reduction_override is not None \
                else kwargs.pop(
                    f'{handler_key}_reduction', None)
            if self.__skip_handler__(
                    loss_handler=handler,
                    loss_weight_override=loss_weight_override):
                continue
            handler_input = init_handler_input.copy()
            handler_input.update({
                'loss_weight_override': loss_weight_override,
                'reduction_override': reduction_override,
            })
            loss_tensor = None
            # e.g. shape prior loss
            if not handler.requires_input():
                loss_tensor = handler(**handler_input)
            # e.g. keypoints 3d mse loss
            else:
                for input_inst in input_list:
                    if input_inst.handler_key == handler_key:
                        handler_input['related_input'] = input_inst
                        loss_tensor = handler(**handler_input)
            # if loss computed, record it in losses
            if loss_tensor is not None:
                losses[handler_key] = loss_tensor

        total_loss = 0
        for key, loss in losses.items():
            if loss.ndim == 3:
                total_loss = total_loss + loss.sum(dim=(2, 1))
            elif loss.ndim == 2:
                total_loss = total_loss + loss.sum(dim=-1)
            else:
                total_loss = total_loss + loss
        losses['total_loss'] = total_loss

        # warn once if there's item still in popped kwargs
        if not self.__stage_kwargs_warned__ and \
                len(kwargs) > 0:
            warn_msg = 'Args below do not match any loss handler:\n'
            table = prettytable.PrettyTable()
            table.field_names = ['Arg key', 'Arg value']
            for key, value in kwargs.items():
                table.add_row([key, value])
            warn_msg = warn_msg + table.get_string()
            self.logger.warning(warn_msg)
            self.__stage_kwargs_warned__ = True

        return losses

    def __match_init_batch_size__(self, init_param: torch.Tensor,
                                  default_param: torch.Tensor,
                                  batch_size: int) -> torch.Tensor:
        """A helper function to ensure body model parameters have the same
        batch size as the input keypoints.

        Args:
            init_param:
                Input initial body model parameters, may be None
            default_param:
                Default parameters if init_param is None. Typically
                it is from the body model.
            batch_size:
                Batch size of input.

        Returns:
            param: body model parameters with batch size aligned
        """

        # param takes init values
        param = init_param.detach().clone() \
            if init_param is not None \
            else default_param.detach().clone()

        # expand batch dimension to match batch size
        param_batch_size = param.shape[0]
        if param_batch_size != batch_size:
            if param_batch_size == 1:
                param = param.repeat(batch_size, *[1] * (param.ndim - 1))
            else:
                self.logger.error(
                    'Init param does not match the batch size of '
                    'keypoints, and is not 1.')
                raise ValueError

        # shape check
        if param.shape[0] != batch_size or \
                param.shape[1:] != default_param.shape[1:]:
            self.logger.error(f'Shape mismatch: {param.shape} vs' +
                              f' {default_param.shape}')
            raise ValueError
        return param

    def __set_keypoint_indexes__(self) -> None:
        """Set keypoint indexes to 1) body parts to be assigned different
        weights 2) be ignored for keypoint loss computation.

        Returns:
            None
        """
        convention = self.body_model.keypoint_convention

        # obtain ignore keypoint indexes
        if self.ignore_keypoints is not None:
            self.ignore_keypoint_idxs = []
            for keypoint_name in self.ignore_keypoints:
                keypoint_idx = get_keypoint_idx(
                    keypoint_name, convention=convention)
                if keypoint_idx != -1:
                    self.ignore_keypoint_idxs.append(keypoint_idx)

        # obtain body part keypoint indexes
        shoulder_keypoint_idxs = get_keypoint_idxs_by_part(
            'shoulder', convention=convention)
        hip_keypoint_idxs = get_keypoint_idxs_by_part(
            'hip', convention=convention)
        self.shoulder_hip_keypoint_idxs = [
            *shoulder_keypoint_idxs, *hip_keypoint_idxs
        ]

    def get_keypoint_weight(self,
                            use_shoulder_hip_only: bool = False,
                            body_weight: float = 1.0) -> torch.Tensor:
        """Get per keypoint weight.

        Args:
            use_shoulder_hip_only (bool, optional):
                Whether to use only shoulder and hip
                keypoints for loss computation. This is useful in the
                warming-up stage to find a reasonably good initialization.
                Defaults to False.
            body_weight (float, optional):
                Weight of body keypoints. Body part segmentation
                definition is included in the HumanData convention.
                Defaults to 1.0.

        Returns:
            torch.Tensor: Per keypoint weight tensor of shape (K).
        """
        n_keypoints = self.body_model.get_joint_number()

        if use_shoulder_hip_only:
            weight = torch.zeros([n_keypoints]).to(self.device)
            weight[self.shoulder_hip_keypoint_idxs] = 1.0
            weight = weight * body_weight
        else:
            weight = torch.ones([n_keypoints]).to(self.device)
            weight = weight * body_weight

        if hasattr(self, 'ignore_keypoint_idxs'):
            weight[self.ignore_keypoint_idxs] = 0.0

        return weight

    def __expand_betas__(self, batch_size, betas):
        """A helper function to expand the betas's first dim to match batch
        size such that the same beta parameters can be used for all frames in a
        video sequence.

        Notes:
            B: batch size
            K: number of keypoints
            D: shape dimension

        Args:
            batch_size: batch size
            betas: shape (B, D)

        Returns:
            betas_video: expanded betas
        """
        # no expansion needed
        if batch_size == betas.shape[0]:
            return betas

        # first dim is 1
        else:
            feat_dim = betas.shape[-1]
            betas_video = betas.view(1, feat_dim).expand(batch_size, feat_dim)

        return betas_video

    def __check_verts_requirement__(self, input_list: List[BaseInput],
                                    **kwargs) -> bool:
        """Check whether vertices are required by loss handlers. For those
        which doesn't need a related input, we take logical or of
        handler.requires_verts(). For those which needs input, we only take
        them when the input is given.

        Args:
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.

        kwargs:
            Stage control keyword arguments, including override weights
            of each loss handler.

        Returns:
            bool: Whether verts are required by any of the handlers.
        """
        any_require_verts = False
        kwargs = kwargs.copy()
        for handler in self.loss_handlers:
            handler_key = handler.handler_key
            loss_weight_override = kwargs.pop(f'{handler_key}_weight', None)
            if self.__skip_handler__(
                    loss_handler=handler,
                    loss_weight_override=loss_weight_override):
                continue
            # e.g. shape prior loss
            if not handler.requires_input():
                any_require_verts = any_require_verts or \
                    handler.requires_verts()
            # e.g. keypoints 3d mse loss
            else:
                for input_inst in input_list:
                    if input_inst.handler_key == handler_key:
                        any_require_verts = any_require_verts or \
                            handler.requires_verts()
        return any_require_verts

    @staticmethod
    def __compute_relative_change__(pre_v, cur_v):
        """Compute relative loss change. If relative change is small enough, we
        can apply early stop to accelerate the optimization. (1) When one of
        the value is larger than 1, we calculate the relative change by diving
        their max value. (2) When both values are smaller than 1, it degrades
        to absolute change. Intuitively, if two values are small and close,
        dividing the difference by the max value may yield a large value.

        Args:
            pre_v: previous value
            cur_v: current value

        Returns:
            float: relative change
        """
        return np.abs(pre_v - cur_v) / max([np.abs(pre_v), np.abs(cur_v), 1])

    @staticmethod
    def __skip_handler__(loss_handler: BaseHandler,
                         loss_weight_override: float) -> bool:
        """Whether to skip loss computation. If loss is None, it will directly
        skip the loss to avoid RuntimeError. If loss is not None, the table
        below shows the return value. If the return value is True, it means the
        computation of loss can be skipped. As the result is 0 even if it is
        calculated, we can skip it to save computational cost.

        | loss.loss_weight | loss_weight_override | returns |
        | ---------------- | -------------------- | ------- |
        |      == 0        |         None         |   True  |
        |      != 0        |         None         |   False |
        |      == 0        |         == 0         |   True  |
        |      != 0        |         == 0         |   True  |
        |      == 0        |         != 0         |   False |
        |      != 0        |         != 0         |   False |

        Args:
            loss_handler(BaseHandler):
                loss_handler is an instance of subclass of BaseHandler.
                loss_handler.get_loss_weight() is assigned
                when loss is initialized.
            loss_weight_override:
                loss_weight_override used to override init loss_weight.

        Returns:
            bool: True means skipping loss computation, and vice versa.
        """
        if loss_handler is None or \
                (loss_handler.get_loss_weight() == 0 and
                 loss_weight_override is None) or \
                loss_weight_override == 0:
            return True
        else:
            return False



import logging
import numpy as np
import prettytable
from typing import Tuple, Union
from xrprimer.utils.log_utils import get_logger


def __init_valid_views_dict__(concerned_n_view: int = 3) -> dict:
    """Create a dict for valid view number statistics。

    Args:
        concerned_n_view (int, optional):
            If a point's valid view number is no greater than
            concerned_n_view, it will be counted.
            Defaults to 3.

    Returns:
        dict
    """
    valid_views_dict = {}
    for n_view in range(concerned_n_view):
        valid_views_dict[n_view] = 0.0
    return valid_views_dict


def get_valid_views_stats(points_mask: np.ndarray,
                          concerned_n_view: int = 3,
                          return_rate: bool = True) -> Tuple[dict, str]:
    """Ignoring masked keypoints, define a pair containing all views about a
    single keypoint, in one frame. Count how many valid views in each pair, and
    print percentage of critical pairs in a table.

    Args:
        points_mask (np.ndarray):
            An ndarray of mask, in shape
            [n_view, n_point, 1].
        concerned_n_view (int, optional):
            If a point's valid view number is no greater than
            concerned_n_view, it will be counted.
            Defaults to 3.
        return_rate (bool, optional):
            Whether to return invalid rate.
            If false, return invalid_pairs count,
            else return invalid_pairs / total_pairs.
            Defaults to True.
    Returns:
        Tuple[dict, str]:
            valid_stats_dict(dict):
                Keys are view number and
                values are pair number/ratio.
            table(str):
                Table of valid_stats_dict
                generated by prettytable.
    """
    # init valid count
    valid_stats_dict = __init_valid_views_dict__(
        concerned_n_view=concerned_n_view)
    total_pairs = points_mask.shape[1]
    for point_index in range(total_pairs):
        pair_data = points_mask[:, point_index, 0]
        nan_data = pair_data[np.isnan(pair_data)]
        # if marked by nan, skip
        if len(nan_data) > 0:
            total_pairs -= 1
            continue
        valid_mask = (pair_data == 1.0)
        # check how many valid views in one data pair
        n_valid = int(np.sum(valid_mask, axis=0))
        # if critical, count it
        if n_valid in valid_stats_dict.keys():
            valid_stats_dict[n_valid] += 1.0
    # get ratio if required
    if return_rate and total_pairs > 0.0:
        for key in valid_stats_dict.keys():
            valid_stats_dict[key] = \
                valid_stats_dict[key] / float(total_pairs)
    table = prettytable.PrettyTable()
    table.field_names = ['Valid Views', 'Pairs']
    for key, item in valid_stats_dict.items():
        table.add_row([key, item])
    table = '\n' + table.get_string()
    return valid_stats_dict, table


def prepare_triangulate_input(
    camera_number: int,
    points: Union[np.ndarray, list, tuple],
    points_mask: Union[np.ndarray, list, tuple] = None,
    logger: Union[None, str, logging.Logger] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Prepare points and points_mask to numpy ndarray. If check fails, raise
    an error.

    Args:
        camera_number (int):
            Number of cameras.
        points (Union[np.ndarray, list, tuple]):
                An ndarray or a nested list of points2d, in shape
                [n_view, ..., 2+n], n >= 0.
                [...] could be [n_keypoints],
                [n_frame, n_keypoints],
                [n_frame, n_person, n_keypoints], etc.
                If length of the last dim is greater
                than 2, the redundant data will be
                concatenated to output, not modified.
        points_mask (Union[np.ndarray, list, tuple], optional):
            An ndarray or a nested list of mask, in shape
            [n_view, ..., 1].
            If points_mask[index] == 1, points[index] is valid
            for triangulation, else it is ignored.
            If points_mask[index] == np.nan, the whole pair will
            be ignored and not counted by any method.
            Defaults to None.
        logger (Union[None, str, logging.Logger], optional):
            Logger for logging. If None, root logger will be selected.
            Defaults to None.

    Raises:
        TypeError: Type of points is not in (np.ndarray, list, tuple).
        TypeError: Type of points_mask is not in (np.ndarray, list, tuple).
        ValueError: View number of input does not match camera_number.
        ValueError: points.shape[-1] must not be fewer than 2.
        ValueError: points_mask must be [n_view, ..., 1] like points.

    Returns:
        Tuple[np.ndarray, np.ndarray]:
            points and points_mask in type ndarray.
    """
    logger = get_logger(logger)
    if isinstance(points, list) or isinstance(points, tuple):
        points = np.asarray(points)
    # check points type
    if not isinstance(points, np.ndarray):
        logger.error('Type of points is not in (np.ndarray, list, tuple).\n' +
                     f'Type: {type(points)}.')
        raise TypeError
    if points_mask is None:
        points_mask = np.ones_like(points[..., 0:1])
    elif isinstance(points_mask, list) or isinstance(points_mask, tuple):
        points_mask = np.asarray(points_mask)
    # check points_mask type
    if not isinstance(points_mask, np.ndarray):
        logger.error(
            'Type of points_mask is not in (np.ndarray, list, tuple).\n' +
            f'Type: {type(points_mask)}.')
        raise TypeError
    if len(points_mask.shape) == len(points.shape) - 1:
        points_mask = np.expand_dims(points_mask, -1)
    # check points shape
    if not (points.shape[0] == points_mask.shape[0]
            and points.shape[0] == camera_number):
        logger.error('View number of input does not' +
                     ' equal to triangulator\'s camera number.\n' +
                     f'points.shape: {points.shape}\n' +
                     f'points_mask.shape: {points_mask.shape}\n' +
                     f'camera number: {camera_number}\n')
        raise ValueError
    if points.shape[-1] < 2:
        logger.error('points.shape[-1] must not be fewer than 2.' +
                     f'points.shape: {points.shape}')
        raise ValueError
    # check points_mask shape
    if points.shape[:-1] != points_mask.shape[:-1] or\
            points_mask.shape[-1] != 1:
        logger.error('points_mask must be [n_view, ..., 1] like points.\n' +
                     f'points_mask.shape: {points_mask.shape}' +
                     f'points.shape: {points.shape}')
        raise ValueError
    return points, points_mask


def parse_keypoints_mask(
    keypoints: Union[np.ndarray, list, tuple],
    keypoints_mask: Union[np.ndarray, list, tuple] = None,
    logger: Union[None, str, logging.Logger] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Get a new selection mask according to keypoints_mask.

    Args:
        keypoints (Union[np.ndarray, list, tuple]):
            An ndarray or a nested list of points2d, in shape
            [n_view, ..., n_keypoints, 2+n].
            [...] could be [n_keypoints],
            [n_frame, n_keypoints],
            [n_frame, n_person, n_keypoints], etc.
            It offers a shape reference for the returned mask.
        keypoints_mask (Union[np.ndarray, list, tuple], optional):
            keypoints_mask in HumanData,
            marking which keypoints is valid.
            Defaults to None.
        logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.

    Raises:
        ValueError:
            Keypoints number of points does not
            match length of keypoints_mask.

    Returns:
        np.ndarray:
            triangulate_mask for points selection and
            triangulation.
    """
    logger = get_logger(logger)
    keypoints, triangulate_mask = prepare_triangulate_input(
        camera_number=keypoints.shape[0], points=keypoints, logger=logger)
    init_points_mask_shape = triangulate_mask.shape
    if keypoints.shape[-2] != keypoints_mask.shape[0]:
        logger.error('Keypoints number of points does not' +
                     ' match length of keypoints_mask.\n' +
                     f'keypoints.shape: {keypoints.shape}' +
                     f'keypoints_mask.shape: {keypoints_mask.shape}')
        raise ValueError
    nan_indexes = np.where(keypoints_mask == 0)
    triangulate_mask = triangulate_mask.reshape(-1, init_points_mask_shape[-2],
                                                init_points_mask_shape[-1])
    triangulate_mask[:, nan_indexes[0], :] = np.nan
    triangulate_mask = triangulate_mask.reshape(*init_points_mask_shape)
    return triangulate_mask



import cv2
import logging
import numpy as np
import torch
from typing import Union

from xrmocap.utils.geometry import get_affine_transform, get_scale
from .base_image_transform import BaseImageTransform


class WarpAffine(BaseImageTransform):

    def __init__(self,
                 image_size: np.ndarray,
                 flag: str = 'inter_linear',
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Warp image array with given affine transformation matrix and given
        image size.

        Args:
            image_size (np.ndarray):
                Output image size.
            flag (str, optional):
                Interpolation method for resizing. Defaults to 'inter_linear'.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        BaseImageTransform.__init__(self, logger=logger)
        flags_dict = {
            'inter_linear': cv2.INTER_LINEAR,
            'inter_nearest': cv2.INTER_NEAREST,
            'inter_area': cv2.INTER_AREA,
            'inter_cubid': cv2.INTER_CUBIC,
            'inter_lanczos4': cv2.INTER_LANCZOS4
        }

        if flag in flags_dict.keys():
            self.flag = flags_dict[flag]
        else:
            self.logger.warning(
                'Unrecognized flag, using cv2.INTER_LINEAR by default.')
            self.flag = cv2.INTER_LINEAR

        self.image_size = image_size

    def forward(self, input: np.ndarray) -> Union[np.ndarray, torch.Tensor]:
        """Forward function of WarpAffine.

        Args:
            input (np.ndarray):
                An array of image with [h, w, n_ch].

        Returns:
            np.ndarray
        """
        height, width = input.shape[0:2]
        c = np.array([width / 2.0, height / 2.0])
        s = get_scale((width, height), self.image_size)
        r = 0  # NOTE: do not apply rotation augmentation
        trans = get_affine_transform(c, s, r, self.image_size, inv=0)
        warped_img = cv2.warpAffine(
            input,
            trans, (int(self.image_size[0]), int(self.image_size[1])),
            flags=self.flag)
        return warped_img



# yapf: disable
from mmcv.utils import Registry

from .bbox_detection.mmdet_detector import MMdetDetector
from .bbox_detection.mmtrack_detector import MMtrackDetector
from .keypoints_estimation.mmpose_top_down_estimator import (
    MMposeTopDownEstimator,
)

# yapf: enable

DETECTORS = Registry('detector')
DETECTORS.register_module(
    name=('MMposeTopDownEstimator'), module=MMposeTopDownEstimator)
DETECTORS.register_module(name=('MMdetDetector'), module=MMdetDetector)
DETECTORS.register_module(name=('MMtrackDetector'), module=MMtrackDetector)


def build_detector(cfg):
    """Build detector."""
    return DETECTORS.build(cfg)



import cv2
import logging
import numpy as np
from tqdm import tqdm
from typing import Union
from xrprimer.utils.ffmpeg_utils import video_to_array
from xrprimer.utils.log_utils import get_logger

try:
    from mmtrack.apis import inference_mot, init_model
    has_mmtrack = True
    import_exception = ''
except (ImportError, ModuleNotFoundError):
    has_mmtrack = False
    import traceback
    stack_str = ''
    for line in traceback.format_stack():
        if 'frozen' not in line:
            stack_str += line + '\n'
    import_exception = traceback.format_exc() + '\n'
    import_exception = stack_str + import_exception


class MMtrackDetector:
    """Detector wrapped from mmtracking."""

    def __init__(self,
                 mmtrack_kwargs: dict,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Init a detector from mmtracking.

        Args:
            mmtrack_kwargs (dict):
                A dict contains args of mmtrack.apis.init_model.
                Necessary keys: config
                Optional keys:
                    checkpoint, device, cfg_options,
                    verbose_init_params.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        self.logger = get_logger(logger)
        if not has_mmtrack:
            self.logger.error(import_exception)
            raise ModuleNotFoundError(
                'Please install mmtrack to run detection with tracking.')
        # build the detector from a config file and a checkpoint file
        self.track_model = init_model(**mmtrack_kwargs)

    def infer_array(self,
                    image_array: Union[np.ndarray, list],
                    disable_tqdm: bool = False,
                    multi_person: bool = False) -> list:
        """Infer frames already in memory(ndarray type).

        Args:
            image_array (Union[np.ndarray, list]):
                BGR image ndarray in shape [n_frame, height, width, 3],
                or a list of image ndarrays in shape [height, width, 3] while
                len(list) == n_frame.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.
            multi_person (bool, optional):
                Whether to allow multi-person detection. If False,
                only the biggest bbox will be returned.
                Defaults to False.

        Returns:
            list:
                List of bboxes. Shape of the nested lists is
                (n_frame, n_human, 5)
                and each bbox is (x, y, x, y, score).
                If someone is missed in one frame, there
                will be a None.
        """
        n_frame = len(image_array)
        mframe_mmtrack_results = []
        for frame_idx in tqdm(range(n_frame), disable=disable_tqdm):
            mmtrack_results = inference_mot(
                self.track_model, image_array[frame_idx], frame_id=frame_idx)
            mframe_mmtrack_results.append(mmtrack_results)

        ret_list = process_mmtrack_results(
            mframe_mmtrack_results=mframe_mmtrack_results,
            multi_person=multi_person)
        return ret_list

    def infer_frames(self,
                     frame_path_list: list,
                     disable_tqdm: bool = False,
                     multi_person: bool = False,
                     load_batch_size: Union[None, int] = None) -> list:
        """Infer frames from file.

        Args:
            frame_path_list (list):
                A list of frames' absolute paths.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.
            multi_person (bool, optional):
                Whether to allow multi-person detection, which is
                slower than single-person.
                Defaults to False.
            load_batch_size (Union[None, int], optional):
                How many frames are loaded at the same time.
                Defaults to None, load all frames in frame_path_list.

        Returns:
            list:
                List of bboxes. Shape of the nested lists is
                (n_frame, n_human, 5)
                and each bbox is (x, y, x, y, score).
                If someone is missed in one frame, there
                will be a None.
        """
        mframe_mmtrack_results = []
        if load_batch_size is None:
            load_batch_size = len(frame_path_list)
        for start_idx in range(0, len(frame_path_list), load_batch_size):
            end_idx = min(len(frame_path_list), start_idx + load_batch_size)
            if load_batch_size < len(frame_path_list):
                self.logger.info(
                    'Processing mmtrack on frames' +
                    f'({start_idx}-{end_idx})/{len(frame_path_list)}')
            image_array_list = []
            for frame_abs_path in frame_path_list[start_idx:end_idx]:
                img_np = cv2.imread(frame_abs_path)
                image_array_list.append(img_np)
            n_frame = len(image_array_list)
            for sub_idx in tqdm(range(n_frame), disable=disable_tqdm):
                mmtrack_results = inference_mot(
                    self.track_model,
                    image_array_list[sub_idx],
                    frame_id=start_idx + sub_idx)
                mframe_mmtrack_results.append(mmtrack_results)
        ret_list = process_mmtrack_results(
            mframe_mmtrack_results=mframe_mmtrack_results,
            multi_person=multi_person)
        return ret_list

    def infer_video(self,
                    video_path: str,
                    disable_tqdm: bool = False,
                    multi_person: bool = False) -> list:
        """Infer frames from a video file.

        Args:
            video_path (str):
                Path to the video to be detected.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.
            multi_person (bool, optional):
                Whether to allow multi-person detection, which is
                slower than single-person.
                Defaults to False.

        Returns:
            list:
                List of bboxes. Shape of the nested lists is
                (n_frame, n_human, 5)
                and each bbox is (x, y, x, y, score).
                If someone is missed in one frame, there
                will be a None.
        """
        image_array = video_to_array(input_path=video_path, logger=self.logger)
        ret_list = self.infer_array(
            image_array=image_array,
            disable_tqdm=disable_tqdm,
            multi_person=multi_person)
        return ret_list


def process_mmtrack_results(mframe_mmtrack_results: list,
                            multi_person: bool = False) -> list:
    """Process output of mmtrack, convert it to bbox list and count frame of
    each tracked human.

    Args:
        mframe_mmtrack_results (list): A list of mmtrack inference output.
        multi_person (bool, optional): Whether to allow multi-person detection,
            which is slower than single-person. Defaults to False.

    Returns:
        list:
            List of bboxes. Shape of the nested lists is
            (n_frame, n_human, 5)
            and each bbox is (x, y, x, y, score).
            If someone is missed in one frame, there
            will be a None.
    """
    # 'track_results' is changed to 'track_bboxes'
    # in https://github.com/open-mmlab/mmtracking/pull/300
    for mmtrack_results in mframe_mmtrack_results:
        if len(mmtrack_results) > 0:
            if 'track_bboxes' in mmtrack_results:
                track_key = 'track_bboxes'
            else:
                track_key = 'track_results'
            break
    mframe_count_dict = {}
    mframe_result_dict = {}
    # record count of each tracked object
    # and bboxes in each frame
    for frame_idx, mmtrack_results in enumerate(mframe_mmtrack_results):
        tracking_results = mmtrack_results[track_key][0]
        sframe_result_dict = {}
        for track in tracking_results:
            track_id = int(track[0])
            sframe_result_dict[track_id] = track[1:]
            if track_id in mframe_count_dict:
                mframe_count_dict[track_id] += 1
            else:
                mframe_count_dict[track_id] = 1
        mframe_result_dict[frame_idx] = sframe_result_dict
    max_track_id = len(mframe_count_dict)
    # prepare the returned bbox list
    ret_list = []
    if multi_person:
        for frame_idx in range(len(mframe_mmtrack_results)):
            frame_bboxes = []
            for track_idx in range(max_track_id):
                if track_idx in mframe_result_dict[frame_idx]:
                    frame_bboxes.append(
                        mframe_result_dict[frame_idx][track_idx])
                else:
                    frame_bboxes.append([
                        0.0,
                    ] * 5)
            ret_list.append(frame_bboxes)
    else:
        # get the object observed in most frames
        longes_track_id = 0
        for key in mframe_count_dict.keys():
            if key > max_track_id:
                max_track_id = key
            if mframe_count_dict[key] > mframe_count_dict[longes_track_id]:
                longes_track_id = key
        for frame_idx in range(len(mframe_mmtrack_results)):
            frame_bboxes = []
            if longes_track_id in mframe_result_dict[frame_idx]:
                frame_bboxes.append(
                    mframe_result_dict[frame_idx][longes_track_id])
            else:
                frame_bboxes.append([
                    0.0,
                ] * 5)
            ret_list.append(frame_bboxes)
    return ret_list



# yapf: disable
from xrmocap.core.hook.smplify_hook.smplify_base_hook import SMPLifyBaseHook
from xrmocap.core.hook.smplify_hook.smplify_verbose_hook import (
    SMPLifyVerboseHook,
)

# yapf: enable

__all__ = ['SMPLifyBaseHook', 'SMPLifyVerboseHook']



import logging
import numpy as np
from typing import Union

from xrmocap.data_structure.keypoints import Keypoints
from .base_optimizer import BaseOptimizer


class TrajectoryOptimizer(BaseOptimizer):

    def __init__(self,
                 n_max_frame: int = 9,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Look for kps3d that deviate from the trajectory, and replace it by
        interpolation.

        Args:
            n_max_frame (int, optional):
                Find the maximum range of valid points. Defaults to 9.
            verbose (bool, optional):
                Whether to log info.
                Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__(verbose=verbose, logger=logger)
        self.n_max_frame = n_max_frame

    def optimize_keypoints3d(self, keypoints3d: Keypoints,
                             **kwargs: dict) -> Keypoints:
        """Forward function of keypoints3d optimizer.

        Args:
            keypoints3d (Keypoints): Input keypoints3d.
        kwargs:
            Redundant keyword arguments to be
            ignored.

        Returns:
            Keypoints: The optimized keypoints3d.
        """
        if keypoints3d.dtype == 'numpy':
            keypoints3d_np = keypoints3d
        else:
            keypoints3d_np = keypoints3d.to_numpy()
            self.logger.warning(
                'TrajectoryOptimizer only support numpy kps for now,' +
                ' the input kps has been converted to numpy.')
        ret_keypoints3d = keypoints3d_np.clone()
        ret_kps3d_arr = ret_keypoints3d.get_keypoints()
        n_person = keypoints3d_np.get_person_number()
        for person_idx in range(n_person):
            kps3d_arr = keypoints3d_np.get_keypoints()[:, person_idx]
            kps3d_mask = keypoints3d_np.get_mask()[:, person_idx]
            optimized_kps3d = self.check_kps3d(kps3d_arr, kps3d_mask)
            ret_kps3d_arr[:, person_idx] = optimized_kps3d
        ret_keypoints3d.set_keypoints(ret_kps3d_arr)
        return ret_keypoints3d

    def check_kps3d(self, kps3d_arr: np.ndarray,
                    kps3d_mask: np.ndarray) -> np.ndarray:
        kps3d = kps3d_arr[..., :3]
        kps3d_score = kps3d_arr[..., 3:4]
        n_frame = kps3d_arr.shape[0]
        n_kps3d = kps3d_arr.shape[1]
        person_nan = np.where(np.sum(kps3d_mask, axis=1) == 0)[0]
        kps3d[person_nan] = np.nan

        for frame_idx in range(2, n_frame - 1):
            for kps3d_idx in range(n_kps3d):
                if np.isnan(kps3d[frame_idx, kps3d_idx]).all():
                    continue
                calc_curr_dist = True
                for i in range(1, self.n_max_frame):
                    if frame_idx - i < 1:
                        break
                    curr_dist = np.linalg.norm(
                        kps3d[frame_idx, kps3d_idx] -
                        kps3d[frame_idx - i, kps3d_idx],
                        ord=2) / i
                    if curr_dist > 0:
                        for j in range(frame_idx - i - 1,
                                       frame_idx - i - self.n_max_frame, -1):
                            if not np.isnan(kps3d[j, kps3d_idx]).all():
                                dist_threshold = 2 * np.linalg.norm(
                                    kps3d[frame_idx - i, kps3d_idx] -
                                    kps3d[j, kps3d_idx],
                                    ord=2) / (
                                        frame_idx - i - j)
                                if curr_dist > dist_threshold:
                                    kps3d[frame_idx, kps3d_idx] = np.nan
                                calc_curr_dist = False
                                break
                    if not calc_curr_dist:
                        break
        kps3d_score[person_nan] = np.nan
        return np.concatenate((kps3d, kps3d_score), axis=-1)



# yapf: disable
from mmcv.utils import Registry
from torchvision.transforms import Normalize, Resize, ToTensor

from .color import BGR2RGB
from .convert import CV2ToPIL
from .load import LoadImageCV2, LoadImagePIL
from .shape import WarpAffine

# yapf: enable

IMAGE_TRANSFORM = Registry('image_transform')
IMAGE_TRANSFORM.register_module(name='BGR2RGB', module=BGR2RGB)
IMAGE_TRANSFORM.register_module(name='LoadImageCV2', module=LoadImageCV2)
IMAGE_TRANSFORM.register_module(name='LoadImagePIL', module=LoadImagePIL)
IMAGE_TRANSFORM.register_module(name='CV2ToPIL', module=CV2ToPIL)
IMAGE_TRANSFORM.register_module(name='Resize', module=Resize)
IMAGE_TRANSFORM.register_module(name='ToTensor', module=ToTensor)
IMAGE_TRANSFORM.register_module(name='Normalize', module=Normalize)
IMAGE_TRANSFORM.register_module(name='WarpAffine', module=WarpAffine)


def build_image_transform(cfg) -> None:
    """Build image_transform."""
    return IMAGE_TRANSFORM.build(cfg)



# yapf: disable
import logging
import numpy as np
import os.path as osp
from tqdm import tqdm
from typing import List, Union
from xrprimer.utils.log_utils import get_logger
from xrprimer.utils.path_utils import prepare_output_path

from xrmocap.core.evaluation.align_keypoints3d import align_keypoints3d
from xrmocap.core.evaluation.metrics import calc_limbs_accuracy, evaluate
from xrmocap.data.data_visualization.builder import (
    BaseDataVisualization, build_data_visualization,
)
from xrmocap.data.dataset.builder import MviewMpersonDataset, build_dataset
from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.ops.bottom_up_association.builder import (
    FourDAGAssociator, build_bottom_up_associator,
)
from xrmocap.transform.convention.keypoints_convention import get_keypoint_num

# yapf: enable


class BottomUpAssociationEvaluation:
    """Bottom-up association evaluation."""

    def __init__(self,
                 output_dir: str,
                 selected_limbs_name: List[List[str]],
                 dataset: Union[dict, MviewMpersonDataset],
                 associator: Union[dict, FourDAGAssociator],
                 additional_limbs_names: List[List[str]] = [],
                 dataset_visualization: Union[None, dict,
                                              BaseDataVisualization] = None,
                 pred_kps3d_convention: str = 'coco',
                 eval_kps3d_convention: str = 'campus',
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Initialization for the class.

        Args:
            output_dir (str): The path to save results.
            selected_limbs_name (List[List[str]]): The name of selected
                limbs in evaluation.
            additional_limbs_names (List[List[str]]):
                Names at both ends of the limb.
            dataset (Union[dict, MviewMpersonDataset])
            associator (Union[dict, MvposeAssociator])
            dataset_visualization
                (Union[None, dict, BaseDataVisualization], optional):
                Defaults to None.
            pred_kps3d_convention (str, optional): Target convention of
                keypoints3d, Defaults to 'coco'.
            eval_kps3d_convention (str, optional): the convention of
                keypoints3d for evaluation, Defaults to 'campus'.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """

        self.output_dir = output_dir
        self.pred_kps3d_convention = pred_kps3d_convention
        self.eval_kps3d_convention = eval_kps3d_convention
        self.additional_limbs_names = additional_limbs_names
        self.selected_limbs_name = selected_limbs_name
        self.logger = get_logger(logger)

        if isinstance(dataset, dict):
            dataset['logger'] = self.logger
            self.dataset = build_dataset(dataset)
        else:
            self.dataset = dataset
        self.n_views = self.dataset.n_views
        if isinstance(associator, dict):
            associator['logger'] = self.logger
            associator['n_views'] = self.n_views
            self.associator = build_bottom_up_associator(associator)
        else:
            self.associator = associator

        if isinstance(dataset_visualization, dict):
            dataset_visualization['logger'] = self.logger
            self.dataset_visualization = build_data_visualization(
                dataset_visualization)
        else:
            self.dataset_visualization = dataset_visualization

    def run(self, overwrite: bool = False):
        prepare_output_path(
            output_path=self.output_dir,
            allowed_suffix='',
            path_type='dir',
            overwrite=overwrite,
            logger=self.logger)
        n_frame = len(self.dataset)
        n_kps = get_keypoint_num(convention=self.pred_kps3d_convention)
        pred_kps3d = np.zeros(shape=(n_frame, 1, n_kps, 4))
        pred_kps2d = np.zeros(shape=(n_frame, 1, self.n_views, n_kps, 3))
        gt_kps3d = None
        max_identity = 0
        end_of_clip_idxs = []
        identities = []
        for frame_idx, frame_item in enumerate(tqdm(self.dataset)):
            _, _, _, _, kps3d, end_of_clip, kps2d, pafs = frame_item
            if end_of_clip:
                end_of_clip_idxs.append(frame_idx)
            fisheye_list = self.dataset.fisheye_params[0]
            # prepare input for associate single frame

            self.associator.set_cameras(fisheye_list)
            predict_keypoints3d, identities, multi_kps2d, _ = \
                self.associator.associate_frame(kps2d, pafs, end_of_clip)
            # save predict kps3d
            for idx, identity in enumerate(identities):
                if identity > max_identity:
                    n_identity = identity - max_identity
                    pred_kps3d = np.concatenate(
                        (pred_kps3d,
                         np.zeros(shape=(n_frame, n_identity, n_kps, 4))),
                        axis=1)
                    pred_kps2d = np.concatenate(
                        (pred_kps2d,
                         np.zeros(
                             shape=(n_frame, n_identity, self.n_views, n_kps,
                                    3))),
                        axis=1)
                    max_identity = identity
                pred_kps3d[frame_idx,
                           identity] = predict_keypoints3d.get_keypoints()[0,
                                                                           idx]
                # prepare 2d associate result
                if identity in multi_kps2d:
                    pred_kps2d[frame_idx, identity] = multi_kps2d[identity]
            # save ground truth kps3d
            if gt_kps3d is None:
                gt_kps3d = kps3d.numpy()[np.newaxis]
            else:
                gt_kps3d = np.concatenate(
                    (gt_kps3d, kps3d.numpy()[np.newaxis]), axis=0)

        pred_keypoints3d = Keypoints(
            dtype='numpy',
            kps=pred_kps3d,
            mask=pred_kps3d[..., -1] > 0,
            convention=self.pred_kps3d_convention,
            logger=self.logger)
        gt_keypoints3d = Keypoints(
            dtype='numpy',
            kps=gt_kps3d,
            mask=gt_kps3d[..., -1] > 0,
            convention=self.dataset.gt_kps3d_convention,
            logger=self.logger)
        mscene_keypoints_paths = []

        # prepare result
        scene_start_idx = 0
        for scene_idx, scene_end_idx in enumerate(end_of_clip_idxs):
            scene_keypoints = pred_keypoints3d.clone()
            kps3d = scene_keypoints.get_keypoints()[
                scene_start_idx:scene_end_idx + 1, ...]
            mask = scene_keypoints.get_mask()[scene_start_idx:scene_end_idx +
                                              1, ...]
            scene_keypoints.set_keypoints(kps3d)
            scene_keypoints.set_mask(mask)
            npz_path = osp.join(self.output_dir,
                                f'scene{scene_idx}_pred_keypoints3d.npz')
            scene_keypoints.dump(npz_path)
            mscene_keypoints_paths.append(npz_path)

            npz_path = osp.join(self.output_dir,
                                f'scene{scene_idx}_associate_keypoints2d')
            associate_kps2d = pred_kps2d[scene_start_idx:scene_end_idx + 1,
                                         ...]
            np.save(npz_path, associate_kps2d)

            scene_start_idx = scene_end_idx + 1

        # evaluation
        pred_keypoints3d_, gt_keypoints3d_, limbs = align_keypoints3d(
            pred_keypoints3d, gt_keypoints3d, self.eval_kps3d_convention,
            self.selected_limbs_name, self.additional_limbs_names)
        _, eval_table = calc_limbs_accuracy(
            pred_keypoints3d_, gt_keypoints3d_, limbs, logger=self.logger)
        self.logger.info('\n' + eval_table.get_string())
        evel_dict = evaluate(
            pred_keypoints3d_,
            gt_keypoints3d_,
            pck_thres=[100, 200],
            logger=self.logger)
        self.logger.info('MPJPE: {:.2f} ± {:.2f} mm'.format(
            evel_dict['mpjpe_mean'], evel_dict['mpjpe_std']))
        self.logger.info(f'PA-MPJPE: {evel_dict["pa_mpjpe_mean"]:.2f} ±'
                         f'{evel_dict["pa_mpjpe_std"]:.2f} mm')
        self.logger.info(f'PCK@100mm: {evel_dict["pck"][100]:.2f} %')
        self.logger.info(f'PCK@200mm: {evel_dict["pck"][200]:.2f} %')

        # visualization
        if self.dataset_visualization is not None:
            self.dataset_visualization.pred_kps3d_paths = \
                mscene_keypoints_paths
            self.dataset_visualization.run(overwrite=overwrite)



from mmcv.utils import Registry

from .aniposelib_optimizer import AniposelibOptimizer
from .base_optimizer import BaseOptimizer
from .fourdag_base_optimizer import FourDAGBaseOptimizer
from .fourdag_optimization import FourDAGOptimizer
from .median_smooth import MedianSmooth
from .nan_interpolation import NanInterpolation
from .smpl_shape_aware_optimizer import SMPLShapeAwareOptimizer
from .trajectory_optimizer import TrajectoryOptimizer

KEYPOINTS3D_OPTIMIZERS = Registry('keypoints3d_optimizer')

KEYPOINTS3D_OPTIMIZERS.register_module(
    name='NanInterpolation', module=NanInterpolation)
KEYPOINTS3D_OPTIMIZERS.register_module(
    name='TrajectoryOptimizer', module=TrajectoryOptimizer)
KEYPOINTS3D_OPTIMIZERS.register_module(
    name='MedianSmooth', module=MedianSmooth)
KEYPOINTS3D_OPTIMIZERS.register_module(
    name='AniposelibOptimizer', module=AniposelibOptimizer)
KEYPOINTS3D_OPTIMIZERS.register_module(
    name='SMPLShapeAwareOptimizer', module=SMPLShapeAwareOptimizer)
KEYPOINTS3D_OPTIMIZERS.register_module(
    name='FourDAGBaseOptimizer', module=FourDAGBaseOptimizer)
KEYPOINTS3D_OPTIMIZERS.register_module(
    name='FourDAGOptimizer', module=FourDAGOptimizer)


def build_keypoints3d_optimizer(cfg) -> BaseOptimizer:
    """Build keypoints3d optimizer."""
    return KEYPOINTS3D_OPTIMIZERS.build(cfg)



# yapf: disable
from xrmocap.core.evaluation.mvp_evaluation import MVPEvaluation
from xrmocap.core.evaluation.top_down_association_evaluation import (
    TopDownAssociationEvaluation,
)

# yapf: enable

__all__ = ['MVPEvaluation', 'TopDownAssociationEvaluation']



import cv2
import logging
import numpy as np
from tqdm import tqdm
from typing import Union
from xrprimer.utils.ffmpeg_utils import video_to_array
from xrprimer.utils.log_utils import get_logger

from xrmocap.transform.bbox import qsort_bbox_list

try:
    from mmdet.apis import inference_detector, init_detector
    has_mmdet = True
    import_exception = ''
except (ImportError, ModuleNotFoundError):
    has_mmdet = False
    import traceback
    stack_str = ''
    for line in traceback.format_stack():
        if 'frozen' not in line:
            stack_str += line + '\n'
    import_exception = traceback.format_exc() + '\n'
    import_exception = stack_str + import_exception


class MMdetDetector:
    """Detector wrapped from mmdetection."""

    def __init__(self,
                 mmdet_kwargs: dict,
                 batch_size: int = 1,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Init a detector from mmdetection.

        Args:
            mmdet_kwargs (dict):
                A dict contains args of mmdet.apis.init_detector.
                Necessary keys: config, checkpoint
                Optional keys: device, cfg_options
            batch_size (int, optional):
                Batch size for inference. Defaults to 1.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        self.logger = get_logger(logger)
        if not has_mmdet:
            self.logger.error(import_exception)
            raise ModuleNotFoundError('Please install mmdet to run detection.')
        # build the detector from a config file and a checkpoint file
        self.det_model = init_detector(**mmdet_kwargs)
        self.batch_size = batch_size

    def infer_array(self,
                    image_array: Union[np.ndarray, list],
                    disable_tqdm: bool = False,
                    multi_person: bool = False) -> list:
        """Infer frames already in memory(ndarray type).

        Args:
            image_array (Union[np.ndarray, list]):
                BGR image ndarray in shape [n_frame, height, width, 3],
                or a list of image ndarrays in shape [height, width, 3] while
                len(list) == n_frame.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.
            multi_person (bool, optional):
                Whether to allow multi-person detection. If False,
                only the biggest bbox will be returned.
                Defaults to False.

        Returns:
            list:
                List of bboxes. Shape of the nested lists is
                (n_frame, n_human, 5)
                and each bbox is (x, y, x, y, score).
        """
        ret_list = []
        bbox_results = []
        n_frame = len(image_array)
        for start_index in tqdm(
                range(0, n_frame, self.batch_size), disable=disable_tqdm):
            end_index = min(n_frame, start_index + self.batch_size)
            img_batch = image_array[start_index:end_index]
            # mmdet 2.16.0 cannot accept batch in ndarray, only list
            if isinstance(img_batch, np.ndarray):
                list_batch = []
                for _, img in enumerate(img_batch):
                    list_batch.append(img)
                img_batch = list_batch
            mmdet_results = inference_detector(self.det_model, img_batch)
            additional_ret = True
            for frame_result in mmdet_results:
                if len(frame_result) != 2:
                    additional_ret = False
                    break
                for bbox_result in frame_result:
                    if isinstance(bbox_result, np.ndarray) and \
                            bbox_result.shape == (0, 5):
                        additional_ret = False
                        break
                if not additional_ret:
                    break
            # For models like HTC
            if additional_ret:
                bbox_results += [i[0] for i in mmdet_results]
            else:
                bbox_results += mmdet_results
        ret_list = process_mmdet_results(
            bbox_results, multi_person=multi_person)
        return ret_list

    def infer_frames(self,
                     frame_path_list: list,
                     disable_tqdm: bool = False,
                     multi_person: bool = False,
                     load_batch_size: Union[None, int] = None) -> list:
        """Infer frames from file.

        Args:
            frame_path_list (list):
                A list of frames' absolute paths.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.
            multi_person (bool, optional):
                Whether to allow multi-person detection, which is
                slower than single-person.
                Defaults to False.
            load_batch_size (Union[None, int], optional):
                How many frames are loaded at the same time.
                Defaults to None, load all frames in frame_path_list.

        Returns:
            list:
                List of bboxes. Shape of the nested lists is
                (n_frame, n_human, 5)
                and each bbox is (x, y, x, y, score).
        """
        ret_list = []
        if load_batch_size is None:
            load_batch_size = len(frame_path_list)
        for start_idx in range(0, len(frame_path_list), load_batch_size):
            end_idx = min(len(frame_path_list), start_idx + load_batch_size)
            if load_batch_size < len(frame_path_list):
                self.logger.info(
                    'Processing mmdet on frames' +
                    f'({start_idx}-{end_idx})/{len(frame_path_list)}')
            image_array_list = []
            for frame_abs_path in frame_path_list[start_idx:end_idx]:
                img_np = cv2.imread(frame_abs_path)
                image_array_list.append(img_np)
            load_batch_result = self.infer_array(
                image_array=image_array_list,
                disable_tqdm=disable_tqdm,
                multi_person=multi_person)
            ret_list += load_batch_result
        return ret_list

    def infer_video(self,
                    video_path: str,
                    disable_tqdm: bool = False,
                    multi_person: bool = False) -> list:
        """Infer frames from a video file.

        Args:
            video_path (str):
                Path to the video to be detected.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.
            multi_person (bool, optional):
                Whether to allow multi-person detection, which is
                slower than single-person.
                Defaults to False.

        Returns:
            list:
                List of bboxes. Shape of the nested lists is
                (n_frame, n_human, 5)
                and each bbox is (x, y, x, y, score).
        """
        image_array = video_to_array(input_path=video_path, logger=self.logger)
        ret_list = self.infer_array(
            image_array=image_array,
            disable_tqdm=disable_tqdm,
            multi_person=multi_person)
        return ret_list


def process_mmdet_results(mmdet_results: list,
                          cat_id: int = 0,
                          multi_person: bool = True) -> list:
    """Process mmdet results, sort bboxes by area in descending order.

    Args:
        mmdet_results (list):
            Result of mmdet.apis.inference_detector
            when the input is a batch.
            Shape of the nested lists is
            (n_frame, n_category, n_human, 5).
        cat_id (int, optional):
            Category ID. This function will only select
            the selected category, and drop the others.
            Defaults to 0, ID of human category.
        multi_person (bool, optional):
            Whether to allow multi-person detection, which is
            slower than single-person. If false, the function
            only assure that the first person of each frame
            has the biggest bbox.
            Defaults to True.

    Returns:
        list:
            A list of detected bounding boxes.
            Shape of the nested lists is
            (n_frame, n_human, 5)
            and each bbox is (x, y, x, y, score).
    """
    ret_list = []
    only_max_arg = not multi_person
    for _, frame_results in enumerate(mmdet_results):
        cat_bboxes = frame_results[cat_id]
        sorted_bbox = qsort_bbox_list(cat_bboxes, only_max_arg)

        if only_max_arg:
            ret_list.append(sorted_bbox[0:1])
        else:
            ret_list.append(sorted_bbox)
    return ret_list



import logging
import numpy as np
from scipy import signal
from typing import Union

from xrmocap.data_structure.keypoints import Keypoints
from .base_optimizer import BaseOptimizer


def median_filter_data(data: np.ndarray, kernel_size: int) -> np.ndarray:
    """Perform a median filter on an ndarray, along the first axis of data.

    Args:
        data (np.ndarray):
            Points data in shape [n_frame, n_point, point_dim].
        kernel_size (int):
            Size of the filter kernel.

    Returns:
        np.ndarray:
            The filterd result.
    """
    ret_data = np.apply_along_axis(
        __median_filter__, 0, data, kernel_size=kernel_size)
    return ret_data


def __median_filter__(data, kernel_size):
    pad_size = kernel_size + int((kernel_size + 1) / 2)
    padded_data = np.pad(data, (pad_size, pad_size), mode='reflect')
    padded_filterd_data = signal.medfilt(padded_data, kernel_size=kernel_size)
    return padded_filterd_data[pad_size:-pad_size]


class MedianSmooth(BaseOptimizer):

    def __init__(self,
                 kernel_size: int = 7,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Perform a median filter on a Keypoints instance.

        Args:
            kernel_size (int, optional):
                Size of the filter kernel. Defaults to 7.
            verbose (bool, optional):
                Whether to log info.
                Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__(verbose=verbose, logger=logger)
        if kernel_size % 2 == 0:
            self.logger.error('kernel_size of MedianFilter should be odd.')
            raise ValueError
        self.kernel_size = kernel_size

    def optimize_keypoints3d(self, keypoints3d: Keypoints,
                             **kwargs: dict) -> Keypoints:
        """Forward function of keypoints3d optimizer.

        Args:
            keypoints3d (Keypoints): Input keypoints3d.
        kwargs:
            Redundant keyword arguments to be
            ignored.

        Returns:
            Keypoints: The optimized keypoints3d.
        """
        if keypoints3d.dtype == 'numpy':
            keypoints3d_np = keypoints3d
        else:
            keypoints3d_np = keypoints3d.to_numpy()
            self.logger.warning(
                'MedianFilter only support numpy kps for now,' +
                ' the input kps has been converted to numpy.')
        ret_keypoints3d = keypoints3d_np.clone()
        ret_kps_arr = ret_keypoints3d.get_keypoints()
        for person_idx in range(keypoints3d_np.get_person_number()):
            kps_arr = keypoints3d_np.get_keypoints()[:, person_idx, ...]
            kps_interp = median_filter_data(
                kps_arr, kernel_size=self.kernel_size)
            ret_kps_arr[:, person_idx, ...] = kps_interp
        ret_keypoints3d.set_keypoints(ret_kps_arr)
        return ret_keypoints3d



# yapf: disable
import numpy as np
from typing import Tuple, Union

from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.data_structure.limbs import Limbs
from xrmocap.transform.convention.keypoints_convention import (
    convert_keypoints, get_keypoint_idx, get_keypoint_names,
    get_keypoints_factory, get_mapping_dict, human_data,
)

# yapf: enable


def get_limbs_from_keypoints(
    keypoints: Keypoints,
    fill_limb_names: bool = False,
    frame_idx: Union[int, None] = None,
    person_idx: Union[int, None] = None,
    keypoints_factory: Union[dict, None] = None,
) -> Limbs:
    """Get an instance of class Limbs, from a Keypoints instance. It searches
    existing limbs in HumanData convention.

    Args:
        keypoints (Keypoints):
            An instance of Keypoints class.
        fill_limb_names (bool, optional):
            Whether to fill connection names. If True,
            names will be get from HUMAN_DATA_LIMB_NAMES.
            Defaults to False.
        frame_idx (Union[int, None], optional):
            If given, limbs.points will be
            set to keypoints[frame_idx, person_idx, :, :].
            Defaults to None, limbs.points will not be set.
        person_idx (Union[int, None], optional):
            If given, limbs.points will be
            set to keypoints[frame_idx, person_idx, :, :].
            Defaults to None, limbs.points will not be set.
        keypoints_factory (Union[dict, None], optional):
            A dict to store all the keypoint conventions.
            Defaults to None, KEYPOINTS_FACTORY will be set.

    Returns:
        Limbs: an instance of class Limbs.
    """
    logger = keypoints.logger
    if keypoints_factory is None:
        keypoints_factory = get_keypoints_factory()
    limbs_source = human_data.HUMAN_DATA_LIMBS_INDEX
    keypoints_np = keypoints.to_numpy()
    # if both frame_idx and person_idx are set
    # take the frame and person
    if frame_idx is not None and person_idx is not None:
        frame_idx = int(frame_idx)
        person_idx = int(person_idx)
        selected_mask = keypoints.get_mask()[frame_idx, person_idx, :]
        selected_keypoints = keypoints_np.get_keypoints()[frame_idx,
                                                          person_idx, ...]
    # if any of [frame_idx, person_idx] not configured
    # use or result of masks
    else:
        if frame_idx is not None or person_idx is not None:
            logger.warning('Either frame_idx or person_idx has not' +
                           ' been set properly, limbs.points will not be set.')
        n_keypoints = keypoints_np.get_keypoints_number()
        flat_mask = keypoints_np.get_mask().reshape(-1, n_keypoints)
        or_mask = np.sign(np.sum(flat_mask, axis=0))
        selected_mask = or_mask
        selected_keypoints = keypoints_np.get_keypoints()[0, 0, ...]
    one_frame_keypoints = Keypoints(
        dtype='numpy',
        kps=selected_keypoints,
        mask=selected_mask,
        convention=keypoints_np.get_convention(),
        logger=keypoints.logger)
    human_data_keypoints = convert_keypoints(
        keypoints=one_frame_keypoints,
        dst='human_data',
        keypoints_factory=keypoints_factory)
    mapping_back = get_mapping_dict(
        src='human_data',
        dst=keypoints.get_convention(),
        keypoints_factory=keypoints_factory)
    mask = human_data_keypoints.get_mask()[0, 0, :]
    connecntions = []
    parts = []
    part_names = []
    connecntion_names = []
    human_data_kps = get_keypoint_names('human_data')
    for part_name, part_limbs in limbs_source.items():
        part_record = []
        for limb in part_limbs:
            # points index defined in human data
            hd_start_idx = limb[0]
            hd_end_idx = limb[1]
            # both points exits
            if mask[hd_start_idx] + mask[hd_end_idx] == 2:
                src_start_idx = mapping_back[hd_start_idx]
                src_end_idx = mapping_back[hd_end_idx]
                connecntions.append([src_start_idx, src_end_idx])
                part_record.append(len(connecntions) - 1)
                if fill_limb_names:
                    hd_start_name = human_data_kps[hd_start_idx]
                    hd_end_name = human_data_kps[hd_end_idx]
                    conn_name, _ = get_limb_name_by_keypoint_names(
                        hd_start_name, hd_end_name)
                    connecntion_names.append(conn_name)
        if len(part_record) > 0:
            parts.append(part_record)
            part_names.append(part_name)
    points = None if person_idx is None or frame_idx is None else\
        one_frame_keypoints.get_keypoints()[frame_idx, person_idx, ...]
    if not (fill_limb_names and len(connecntion_names) > 0):
        connecntion_names = None
    ret_limbs = Limbs(
        connections=np.asarray(connecntions),
        parts=parts,
        part_names=part_names,
        points=points,
        connection_names=connecntion_names,
        logger=keypoints.logger)
    return ret_limbs


def search_limbs(data_source: str,
                 mask: Union[np.ndarray, tuple, list] = None,
                 keypoints_factory: dict = None) -> Tuple[dict, dict]:
    """Search the corresponding limbs following the basis human_data limbs.

    The
    mask could mask out the incorrect keypoints.
    Args:
        data_source (str): data source type.
        mask (Optional[Union[np.ndarray, tuple, list]], optional):
            refer to keypoints_mapping. Defaults to None.
        keypoints_factory (dict, optional): Dict of all the conventions.
            Defaults to KEYPOINTS_FACTORY.
    Returns:
        Tuple[dict, dict]: (limbs_target, limbs_palette).
    """
    if keypoints_factory is None:
        keypoints_factory = get_keypoints_factory()
    limbs_source = human_data.HUMAN_DATA_LIMBS_INDEX
    limbs_palette = human_data.HUMAN_DATA_PALETTE
    keypoints_source = keypoints_factory['human_data']
    keypoints_target = keypoints_factory[data_source]
    limbs_target = {}
    for k, part_limbs in limbs_source.items():
        limbs_target[k] = []
        for limb in part_limbs:
            flag = False
            if (keypoints_source[limb[0]]
                    in keypoints_target) and (keypoints_source[limb[1]]
                                              in keypoints_target):
                if mask is not None:
                    if mask[keypoints_target.index(keypoints_source[
                            limb[0]])] != 0 and mask[keypoints_target.index(
                                keypoints_source[limb[1]])] != 0:
                        flag = True
                else:
                    flag = True
                if flag:
                    limbs_target.setdefault(k, []).append([
                        keypoints_target.index(keypoints_source[limb[0]]),
                        keypoints_target.index(keypoints_source[limb[1]])
                    ])
        if k in limbs_target:
            if k == 'body':
                np.random.seed(0)
                limbs_palette[k] = np.random.randint(
                    0, high=255, size=(len(limbs_target[k]), 3))
            else:
                limbs_palette[k] = np.array(limbs_palette[k])
    return limbs_target, limbs_palette


def get_limb_name_by_keypoint_names(keypoint_name_0: str,
                                    keypoint_name_1: str) -> Tuple[str, bool]:
    """Search whether the name of a limb connecting keypoint_0 and keypoint_1
    exists in human_data.HUMAN_DATA_LIMB_NAMES. Return a name and the search
    result. If fould, return the preset name, else return
    f'{keypoint_name_a}_to_{keypoint_name_b}', while index(keypoint_name_a) <
    index(keypoint_name_b).

    Args:
        keypoint_name_0 (str):
            Name of one keypoint.
        keypoint_name_1 (str):
            Name of another keypoint.

    Returns:
        Tuple[str, bool]:
            Name of the limb, and whether the name comes from
            preset-HUMAN_DATA_LIMB_NAMES.
    """
    name_dict = human_data.HUMAN_DATA_LIMB_NAMES
    if keypoint_name_0 in name_dict and\
            keypoint_name_1 in name_dict[keypoint_name_0]:
        fould_flag = True
        limb_name = name_dict[keypoint_name_0][keypoint_name_1]
    elif keypoint_name_1 in name_dict and\
            keypoint_name_0 in name_dict[keypoint_name_1]:
        fould_flag = True
        limb_name = name_dict[keypoint_name_1][keypoint_name_0]
    else:
        fould_flag = False
        keypoint_idx_0 = get_keypoint_idx(
            name=keypoint_name_0, convention='human_data')
        keypoint_idx_1 = get_keypoint_idx(
            name=keypoint_name_1, convention='human_data')
        if keypoint_idx_0 < keypoint_idx_1:
            limb_name = f'{keypoint_name_0}_to_{keypoint_name_1}'
        else:
            limb_name = f'{keypoint_name_1}_to_{keypoint_name_0}'
    return limb_name, fould_flag



# yapf: disable
import logging
import numpy as np
from typing import Union

from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.ops.triangulation.builder import (
    AniposelibTriangulator, build_triangulator,
)
from xrmocap.transform.limbs import get_limbs_from_keypoints
from .base_optimizer import BaseOptimizer

# yapf: enable


class AniposelibOptimizer(BaseOptimizer):

    def __init__(self,
                 triangulator: Union[AniposelibTriangulator, dict],
                 smooth_weight: float = 4.0,
                 bone_length_weight: float = 2.0,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Call optim_points in aniposelib to apply smooth and constraint on a
        keypoints3d instance.

        Args:
            triangulator (Union[AniposelibTriangulator, dict]):
                Triangulator or config dict.
            smooth_weight (float, optional):
                Weight of smooth optimization.
                Defaults to 4.0.
            bone_length_weight (float, optional):
                Weight of bone length optimization.
                Defaults to 2.0.
            verbose (bool, optional):
                Whether to log info.
                Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__(verbose=verbose, logger=logger)
        self.smooth_weight = smooth_weight
        self.bone_length_weight = bone_length_weight
        if isinstance(triangulator, dict):
            self.triangulator = build_triangulator(triangulator)
        else:
            self.triangulator = triangulator

    def optimize_keypoints3d(self, keypoints3d: Keypoints,
                             mview_kps2d: np.ndarray,
                             mview_kps2d_mask: np.ndarray,
                             **kwargs: dict) -> Keypoints:
        """Forward function of keypoints3d optimizer.

        Args:
            keypoints3d (Keypoints): Input keypoints3d.
            mview_kps2d (np.ndarray):
                Multi-view keypoints2d array for triangulation,
                in shape [n_view, n_frame, n_person, n_kps, 2+n],
                n >= 0.
            mview_kps2d_mask (np.ndarray):
                Multi-view keypoints2d mask for triangulation,
                in shape [n_view, n_frame, n_person, n_kps, 1].

        kwargs:
            Redundant keyword arguments to be
            ignored.

        Returns:
            Keypoints: The optimized keypoints3d.
        """
        if keypoints3d.dtype == 'numpy':
            keypoints3d_np = keypoints3d
        else:
            keypoints3d_np = keypoints3d.to_numpy()
            self.logger.warning('AniposelibOptimizer only support numpy kps,' +
                                ' the input kps has been converted to numpy.')
        limbs = get_limbs_from_keypoints(keypoints3d)
        n_person = keypoints3d.get_person_number()
        kps3d_src = keypoints3d.get_keypoints()
        kps3d_dst = kps3d_src.copy()
        ignore_idxs = np.where(mview_kps2d_mask != 1)
        mview_kps2d[ignore_idxs[0], ignore_idxs[1], ignore_idxs[2],
                    ignore_idxs[3], :] = np.nan
        connections = limbs.get_connections()
        camera_group = self.triangulator.__prepare_aniposelib_camera__()
        for person_idx in range(n_person):
            kps2d = mview_kps2d[:, :, person_idx, :, :2]
            kps3d = kps3d_src[:, person_idx, :, :3]
            kps3d_dst[:, person_idx, :, :3] = camera_group.optim_points(
                kps2d,
                kps3d,
                scale_smooth=self.smooth_weight,
                scale_length=self.bone_length_weight,
                constraints=connections,
                verbose=self.verbose,
            )
        ret_keypoints3d = keypoints3d_np.clone()
        ret_keypoints3d.set_keypoints(kps3d_dst)
        return ret_keypoints3d



import logging
import numpy as np
from typing import Union

from xrmocap.data_structure.keypoints import Keypoints
from .base_optimizer import BaseOptimizer


def interpolate_np_data(data: np.ndarray) -> np.ndarray:
    """Interpolate data in type ndarray, nan value will be set by
    interpolation.

    Args:
        data (np.ndarray):
            Points data in shape [n_frame, n_point, point_dim].

    Returns:
        np.ndarray:
            The interpolation result.
    """
    ret_data = np.apply_along_axis(__interpolate_np_nan__, 0, data)
    return ret_data


def __interpolate_np_nan__(data):
    # True if nan, False otherwise
    nan_mask = np.isnan(data)
    ret_data = np.copy(data)
    try:
        ret_data[nan_mask] = \
            np.interp(
                np.nonzero(nan_mask)[0],
                np.nonzero(~nan_mask)[0],
                data[~nan_mask])
    except ValueError:
        pass
    return ret_data


def count_masked_nan(points: np.ndarray, mask: np.ndarray) -> int:
    """Count how many points are nan after the mask applied.

    Args:
        points (np.ndarray):
            In shape [frame_n, person_n, dim].
        mask (np.ndarray):
            In shape [frame_n, person_n].

    Returns:
        int: number of np.nan whose mask is 1.
    """
    squeezed_points = np.sum(points, axis=-1, keepdims=False)
    count = np.count_nonzero(
        np.logical_and(np.isnan(squeezed_points), mask != 0))
    return count


class NanInterpolation(BaseOptimizer):

    def __init__(self,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Assign keypoints3d values by interpolation, replace nan points.

        Args:
            verbose (bool, optional):
                Whether to log info.
                Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__(verbose=verbose, logger=logger)

    def optimize_keypoints3d(self, keypoints3d: Keypoints,
                             **kwargs: dict) -> Keypoints:
        """Forward function of keypoints3d optimizer.

        Args:
            keypoints3d (Keypoints): Input keypoints3d.
        kwargs:
            Redundant keyword arguments to be
            ignored.

        Returns:
            Keypoints: The optimized keypoints3d.
        """
        if keypoints3d.dtype == 'numpy':
            keypoints3d_np = keypoints3d
        else:
            keypoints3d_np = keypoints3d.to_numpy()
            self.logger.warning(
                'NanInterpolation only support numpy kps for now,' +
                ' the input kps has been converted to numpy.')
        total_nan_count = 0
        interp_nan_count = 0
        ret_keypoints3d = keypoints3d_np.clone()
        ret_kps_arr = ret_keypoints3d.get_keypoints()
        for person_idx in range(keypoints3d_np.get_person_number()):
            kps_arr = keypoints3d_np.get_keypoints()[:, person_idx, ...]
            mask = keypoints3d_np.get_mask()[:, person_idx, ...]
            kps_interp = interpolate_np_data(kps_arr)
            ret_kps_arr[:, person_idx, ...] = kps_interp
            # record nan
            input_nan_count = count_masked_nan(kps_arr, mask)
            output_nan_count = count_masked_nan(kps_interp, mask)
            total_nan_count += input_nan_count
            interp_nan_count += input_nan_count - output_nan_count
        ret_keypoints3d.set_keypoints(ret_kps_arr)
        if self.verbose:
            self.logger.info(
                f'\nHow many nans are found after mask: {total_nan_count}' +
                f'\nHow many nans are interpolated: {interp_nan_count}')
        return ret_keypoints3d



from mmcv.utils import Registry

from .trainer import MVPTrainer

TRAINERS = Registry('trainers')

TRAINERS.register_module(name='MVPTrainer', module=MVPTrainer)


def build_trainer(cfg):
    """Build registrant."""
    return TRAINERS.build(cfg)



import mmcv
import os
import os.path as osp
import pickle
import shutil
import tempfile
import time
import torch
import torch.distributed as dist
from mmcv.runner import get_dist_info
from xrprimer.utils.log_utils import get_logger


def time_synchronized():
    torch.cuda.synchronize() if torch.cuda.is_available() else None
    return time.time()


def setup_for_distributed(is_master):
    """This function disables printing when not in master process."""
    import builtins as __builtin__
    builtin_print = __builtin__.print

    def print(*args, **kwargs):
        force = kwargs.pop('force', False)
        if is_master or force:
            builtin_print(*args, **kwargs)

    __builtin__.print = print


def init_distributed_mode(world_size, dist_url, logger):
    """This function initiate the distributed training setup, and return the
    current gpu index."""
    logger = get_logger(logger)
    if 'RANK' in os.environ and 'WORLD_SIZE' in os.environ:
        rank = int(os.environ['RANK'])
        world_size = int(os.environ['WORLD_SIZE'])
        gpu_idx = int(os.environ['LOCAL_RANK'])
    elif 'SLURM_PROCID' in os.environ:
        rank = int(os.environ['SLURM_PROCID'])
        gpu_idx = rank % torch.cuda.device_count()
    else:
        distributed = False
        return distributed

    distributed = True

    torch.cuda.set_device(gpu_idx)
    dist_backend = 'nccl'
    logger.info(f'| distributed init (rank {rank}):' f'{dist_url}')
    torch.distributed.init_process_group(
        backend=dist_backend,
        init_method=dist_url,
        world_size=world_size,
        rank=rank)
    torch.distributed.barrier()
    setup_for_distributed(rank == 0)

    return distributed, gpu_idx


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def get_world_size():
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()


def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()


def is_main_process():
    return get_rank() == 0


def collect_results(result_part, size, tmpdir=None):
    rank, world_size = get_dist_info()
    # create a tmp dir if it is not specified
    if tmpdir is None:
        MAX_LEN = 512
        # 32 is whitespace
        dir_tensor = torch.full((MAX_LEN, ),
                                32,
                                dtype=torch.uint8,
                                device='cuda')
        if rank == 0:
            tmpdir = tempfile.mkdtemp()
            tmpdir = torch.tensor(
                bytearray(tmpdir.encode()), dtype=torch.uint8, device='cuda')
            dir_tensor[:len(tmpdir)] = tmpdir
        dist.broadcast(dir_tensor, 0)
        tmpdir = dir_tensor.cpu().numpy().tobytes().decode().rstrip()
    else:
        mmcv.mkdir_or_exist(tmpdir)
    # dump the part result to the dir
    mmcv.dump(result_part, osp.join(tmpdir, f'part_{rank}.pkl'))
    dist.barrier()
    # collect all parts
    if rank != 0:
        return None
    else:
        # load results of all parts from tmp dir
        part_list = []
        for i in range(world_size):
            part_file = osp.join(tmpdir, f'part_{i}.pkl')
            part_list.append(mmcv.load(part_file))
        # sort the results
        ordered_results = []
        for res in zip(*part_list):
            ordered_results.extend(list(res))
        # the dataloader may pad some samples
        ordered_results = ordered_results[:size]
        # remove tmp dir
        shutil.rmtree(tmpdir)
        return ordered_results


def all_gather(data):
    """
    Run all_gather on arbitrary picklable data (not necessarily tensors)
    Args:
        data:
            Any picklable object
    Returns:
        data_list(list):
            List of data gathered from each rank
    """
    world_size = get_world_size()
    if world_size == 1:
        return [data]

    # serialized to a Tensor
    buffer = pickle.dumps(data)
    storage = torch.ByteStorage.from_buffer(buffer)
    tensor = torch.ByteTensor(storage).to('cuda')

    # obtain Tensor size of each rank
    local_size = torch.tensor([tensor.numel()], device='cuda')
    size_list = [torch.tensor([0], device='cuda') for _ in range(world_size)]
    dist.all_gather(size_list, local_size)
    size_list = [int(size.item()) for size in size_list]
    max_size = max(size_list)

    # receiving Tensor from all ranks
    # we pad the tensor because torch all_gather does not support
    # gathering tensors of different shapes
    tensor_list = []
    for _ in size_list:
        tensor_list.append(
            torch.empty((max_size, ), dtype=torch.uint8, device='cuda'))
    if local_size != max_size:
        padding = torch.empty(
            size=(max_size - local_size, ), dtype=torch.uint8, device='cuda')
        tensor = torch.cat((tensor, padding), dim=0)
    dist.all_gather(tensor_list, tensor)

    data_list = []
    for size, tensor in zip(size_list, tensor_list):
        buffer = tensor.cpu().numpy().tobytes()[:size]
        data_list.append(pickle.loads(buffer))

    return data_list



# yapf: disable

import logging
import numpy as np
import torch
from typing import Any, Union

from .smpl_data import SMPLData

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# yapf: enable


class SMPLXData(SMPLData):
    BODY_POSE_LEN = 21
    HAND_POSE_LEN = 15
    JAW_POSE_LEN = 1
    EYE_POSE_LEN = 1
    DEFAULT_BODY_JOINTS_NUM = 144
    BODY_POSE_KEYS = {
        'global_orient',
        'body_pose',
    }
    FULL_POSE_KEYS = {
        'global_orient', 'body_pose', 'left_hand_pose', 'right_hand_pose',
        'jaw_pose', 'leye_pose', 'reye_pose'
    }

    def __init__(self,
                 src_dict: dict = None,
                 gender: Union[Literal['female', 'male', 'neutral'],
                               None] = None,
                 fullpose: Union[np.ndarray, torch.Tensor, None] = None,
                 transl: Union[np.ndarray, torch.Tensor, None] = None,
                 betas: Union[np.ndarray, torch.Tensor, None] = None,
                 expression: Union[np.ndarray, torch.Tensor, None] = None,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Construct a SMPLXData instance with pre-set values. If any of
        gender, fullpose, transl, betas is provided, it will override the item
        in source_dict.

        Args:
            src_dict (dict, optional):
                A dict with items in HumanData fashion.
                Defaults to None.
            gender (Union[
                    Literal['female', 'male', 'neutral'], None], optional):
                Gender of the body model.
                Should be one among ["female", "male", "neutral"].
                Defaults to None.
            fullpose (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for fullpose, in shape [frame_num, 24, 3].
                Defaults to None, zero-tensor will be created.
            transl (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for translation, in shape [frame_num, 3].
                Defaults to None, zero-tensor will be created.
            betas (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for betas,
                in shape [frame_num, betas_dim].
                Defaults to None,
                zero-tensor in shape [frame_num, 10] will be created.
            expression (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for expression,
                in shape [frame_num, expression_dim].
                Defaults to None,
                zero-tensor in shape [frame_num, 10] will be created.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        SMPLData.__init__(
            self,
            src_dict=src_dict,
            gender=gender,
            transl=transl,
            fullpose=fullpose,
            betas=betas,
            logger=logger)
        if expression is None and 'expression' not in self:
            expression = np.zeros(shape=(self.get_batch_size(), 10))
        if expression is not None:
            self.set_expression(expression)
        self.body_joints_num = self.__class__.DEFAULT_BODY_JOINTS_NUM

    @classmethod
    def get_fullpose_dim(cls) -> int:
        """Get dimension of full pose.

        Returns:
            int:
                Dim value. Full pose shall be
                in shape (frame_n, dim, 3)
        """
        global_orient_dim = 1
        ret_sum = global_orient_dim + cls.BODY_POSE_LEN + \
            cls.JAW_POSE_LEN + 2 * cls.EYE_POSE_LEN + \
            2 * cls.HAND_POSE_LEN
        return ret_sum

    def set_expression(self, expression: Union[np.ndarray,
                                               torch.Tensor]) -> None:
        """Set expression data.

        Args:
            expression (Union[np.ndarray, torch.Tensor]):
                Expression parameters in ndarray or tensor,
                in shape [batch_size, n].
                n stands for any positive int, typically it's 10.

        Raises:
            TypeError: Type of expression is not correct.
        """
        if isinstance(expression, torch.Tensor):
            expression = expression.detach().cpu().numpy()
        elif not isinstance(expression, np.ndarray):
            self.logger.error('Type of expression is not correct.\n' +
                              f'Type: {type(expression)}.')
            raise TypeError
        if len(expression.shape) == 1:
            expression = expression[np.newaxis, ...]
        expression_dim = expression.shape[-1]
        expression_np = expression.reshape(-1, expression_dim)
        dict.__setitem__(self, 'expression', expression_np)

    def get_expression(self, repeat_expression: bool = True) -> np.ndarray:
        """Get expression.

        Args:
            repeat_expression (bool, optional):
                Whether to repeat expression when its first dim doesn't match
                batch_size. Defaults to True.

        Returns:
            ndarray:
                expression in shape [batch_size, expression_dims] or
                [1, expression_dims].
        """
        batch_size = self.get_global_orient().shape[0]
        expression = self.__getitem__('expression')
        if repeat_expression and\
                expression.shape[0] == 1 and\
                expression.shape[0] != batch_size:
            expression = expression.repeat(repeats=batch_size, axis=0)
        return expression

    def __setitem__(self, __k: Any, __v: Any) -> None:
        """Set item according to its key.

        Args:
            __k (Any): Key in dict.
            __v (Any): Value in dict.
        """
        if __k == 'expression':
            self.set_expression(__v)
        else:
            SMPLData.__setitem__(self, __k, __v)

    def from_param_dict(self, smplx_dict: dict) -> None:
        """Load SMPLX parameters from smplx_dict.

        Args:
            smplx_dict (dict):
                A dict of ndarray|Tensor parameters.
                global_orient and body_pose are necessary,
                jaw_pose, leye_pose, reye_pose,
                left_hand_pose, right_hand_pose, transl and
                betas are optional.
                Other keys are ignored.

        Raises:
            KeyError: missing necessary keys.
        """
        necessary_keys = {'global_orient', 'body_pose'}
        if not necessary_keys.issubset(smplx_dict):
            self.logger.error('Keys are not enough.\n' +
                              f'smplx_dict\'s keys: {smplx_dict.keys()}')
            raise KeyError
        global_orient = smplx_dict['global_orient']
        batch_size = global_orient.shape[0] \
            if len(global_orient.shape) > 1 else 1
        global_orient = smplx_dict['global_orient'].reshape(batch_size, 3)
        body_pose = smplx_dict['body_pose'].reshape(batch_size, -1)
        if isinstance(global_orient, torch.Tensor):

            def concat_func(data_list, dim):
                return torch.cat(data_list, dim=dim)

            def zeros_func(shape, ref_data):
                return torch.zeros(
                    size=shape, dtype=ref_data.dtype, device=ref_data.device)

        elif isinstance(global_orient, np.ndarray):

            def concat_func(data_list, dim):
                return np.concatenate(data_list, axis=dim)

            def zeros_func(shape, ref_data):
                return np.zeros(shape=shape, dtype=ref_data.dtype)

        jaw_pose = smplx_dict['jaw_pose'].reshape(batch_size, -1) \
            if 'jaw_pose' in smplx_dict else\
            zeros_func(
                shape=[batch_size, self.__class__.JAW_POSE_LEN*3],
                ref_data=global_orient)
        leye_pose = smplx_dict['leye_pose'].reshape(batch_size, -1) \
            if 'leye_pose' in smplx_dict else\
            zeros_func(
                shape=[batch_size, self.__class__.EYE_POSE_LEN*3],
                ref_data=global_orient)
        reye_pose = smplx_dict['reye_pose'].reshape(batch_size, -1) \
            if 'reye_pose' in smplx_dict else\
            zeros_func(
                shape=[batch_size, self.__class__.EYE_POSE_LEN*3],
                ref_data=global_orient)
        if 'left_hand_pose' in smplx_dict:
            if smplx_dict['left_hand_pose'].reshape(
                    batch_size, -1, 3).shape[1] == \
                        self.__class__.HAND_POSE_LEN:
                left_hand_pose = smplx_dict['left_hand_pose'].reshape(
                    batch_size, -1)
            else:
                left_hand_pose = zeros_func(
                    shape=[batch_size, self.__class__.HAND_POSE_LEN * 3],
                    ref_data=global_orient)
                self.logger.warning(
                    'SMPLX is using pca for hands,' +
                    ' left_hand_pose in SMPLXData will be set to zeros.')
        else:
            left_hand_pose = zeros_func(
                shape=[batch_size, self.__class__.HAND_POSE_LEN * 3],
                ref_data=global_orient)
        if 'right_hand_pose' in smplx_dict:
            if smplx_dict['right_hand_pose'].reshape(
                    batch_size, -1, 3).shape[1] == \
                        self.__class__.HAND_POSE_LEN:
                right_hand_pose = smplx_dict['right_hand_pose'].reshape(
                    batch_size, -1)
            else:
                right_hand_pose = zeros_func(
                    shape=[batch_size, self.__class__.HAND_POSE_LEN * 3],
                    ref_data=global_orient)
                self.logger.warning(
                    'SMPLX is using pca for hands,' +
                    ' right_hand_pose in SMPLXData will be set to zeros.')
        else:
            right_hand_pose = zeros_func(
                shape=[batch_size, self.__class__.HAND_POSE_LEN * 3],
                ref_data=global_orient)

        fullpose = concat_func([
            global_orient, body_pose, jaw_pose, leye_pose, reye_pose,
            left_hand_pose, right_hand_pose
        ],
                               dim=1)
        self.set_fullpose(fullpose)
        if 'transl' in smplx_dict:
            self.set_transl(smplx_dict['transl'])
        if 'betas' in smplx_dict:
            self.set_betas(smplx_dict['betas'])

    def to_param_dict(self,
                      repeat_betas: bool = True,
                      repeat_expression: bool = True) -> dict:
        """Split fullpose into global_orient, body_pose, jaw_pose, leye_pose,
        reye_pose, left_hand_pose, right_hand_pose, return all the necessary
        parameters in one dict.

        Args:
            repeat_betas (bool, optional):
                Whether to repeat betas when its first dim doesn't match
                batch_size. Defaults to True.
            repeat_expression (bool, optional):
                Whether to repeat expression when its first dim doesn't match
                batch_size. Defaults to True.

        Returns:
            dict:
                A dict of SMPLX data, whose keys are
                betas, global_orient, transl, global_orient, body_pose,
                jaw_pose, leye_pose, reye_pose, left_hand_pose,
                right_hand_pose, expression.
        """
        dict_to_return = SMPLData.to_param_dict(
            self, repeat_betas=repeat_betas)
        dict_to_return.pop('body_pose')
        fullpose = self.get_fullpose()
        start_idx = 1
        body_pose = fullpose[:, start_idx:start_idx +
                             self.__class__.BODY_POSE_LEN].reshape(
                                 self.get_batch_size(), -1)
        start_idx += self.__class__.BODY_POSE_LEN
        jaw_pose = fullpose[:, start_idx:start_idx +
                            self.__class__.JAW_POSE_LEN].reshape(-1, 3)
        start_idx += self.__class__.JAW_POSE_LEN
        leye_pose = fullpose[:, start_idx:start_idx +
                             self.__class__.EYE_POSE_LEN].reshape(-1, 3)
        start_idx += self.__class__.EYE_POSE_LEN
        reye_pose = fullpose[:, start_idx:start_idx +
                             self.__class__.EYE_POSE_LEN].reshape(-1, 3)
        start_idx += self.__class__.EYE_POSE_LEN
        left_hand_pose = fullpose[:, start_idx:start_idx +
                                  self.__class__.HAND_POSE_LEN].reshape(-1, 3)
        start_idx += self.__class__.HAND_POSE_LEN
        right_hand_pose = fullpose[:, start_idx:start_idx +
                                   self.__class__.HAND_POSE_LEN].reshape(
                                       -1, 3)
        expression = self.get_expression(repeat_expression=repeat_expression)
        dict_to_return.update({
            'body_pose': body_pose,
            'jaw_pose': jaw_pose,
            'leye_pose': leye_pose,
            'reye_pose': reye_pose,
            'left_hand_pose': left_hand_pose,
            'right_hand_pose': right_hand_pose,
            'expression': expression
        })
        return dict_to_return

    def to_tensor_dict(self,
                       repeat_betas: bool = True,
                       repeat_expression: bool = True,
                       device: Union[torch.device, str] = 'cpu') -> dict:
        """It is almost same as self.to_param_dict, but all the values are
        tensors in a specified device. Split fullpose into global_orient and
        body_pose, return all the necessary parameters in one dict.

        Args:
            repeat_betas (bool, optional):
                Whether to repeat betas when its first dim doesn't match
                batch_size. Defaults to True.
            repeat_expression (bool, optional):
                Whether to repeat expression when its first dim doesn't match
                batch_size. Defaults to True.
            device (Union[torch.device, str], optional):
                A specified device. Defaults to CPU_DEVICE. Defaults to 'cpu'.

        Returns:
            dict: A dict of SMPLX data, whose keys are
                betas, body_pose, global_orient and transl, etc.
        """
        np_dict = self.to_param_dict(
            repeat_betas=repeat_betas, repeat_expression=repeat_expression)
        dict_to_return = {}
        for key, value in np_dict.items():
            dict_to_return[key] = torch.tensor(
                value, device=device, dtype=torch.float32)
        return dict_to_return



# yapf: disable
import csv
import logging
import numpy as np
import os.path as osp
import time
from prettytable import PrettyTable
from tqdm import tqdm
from typing import List, Tuple, Union
from xrprimer.utils.log_utils import get_logger
from xrprimer.utils.path_utils import prepare_output_path

from xrmocap.data.data_visualization.builder import (
    BaseDataVisualization, build_data_visualization,
)
from xrmocap.data.dataset.builder import MviewMpersonDataset, build_dataset
from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.ops.top_down_association.builder import (
    MvposeAssociator, build_top_down_associator,
)
from xrmocap.transform.convention.keypoints_convention import (
    convert_keypoints, get_keypoint_idx, get_keypoint_num,
)
from xrmocap.transform.limbs import get_limbs_from_keypoints
from xrmocap.utils.geometry import compute_similarity_transform
from xrmocap.utils.mvpose_utils import (
    add_campus_jaw_headtop, add_campus_jaw_headtop_mask, check_limb_is_correct,
    compute_mpjpe, vectorize_distance,
)

# yapf: enable


class TopDownAssociationEvaluation:
    """Top-down association evaluation."""

    def __init__(self,
                 output_dir: str,
                 selected_limbs_name: List[List[str]],
                 additional_limbs_names: List[List[str]],
                 dataset: Union[dict, MviewMpersonDataset],
                 associator: Union[dict, MvposeAssociator],
                 dataset_visualization: Union[None, dict,
                                              BaseDataVisualization] = None,
                 pred_kps3d_convention: str = 'coco',
                 eval_kps3d_convention: str = 'campus',
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Initialization for the class.

        Args:
            output_dir (str): The path to save results.
            selected_limbs_name (List[List[str]]): The name of selected
                limbs in evaluation.
            additional_limbs_names (List[List[str]]):
                Names at both ends of the limb.
            dataset (Union[dict, MviewMpersonDataset])
            associator (Union[dict, MvposeAssociator])
            dataset_visualization
                (Union[None, dict, BaseDataVisualization], optional):
                Defaults to None.
            pred_kps3d_convention (str, optional): Target convention of
                keypoints3d, Defaults to 'coco'.
            eval_kps3d_convention (str, optional): the convention of
                keypoints3d for evaluation, Defaults to 'campus'.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """

        self.output_dir = output_dir
        self.pred_kps3d_convention = pred_kps3d_convention
        self.eval_kps3d_convention = eval_kps3d_convention
        self.additional_limbs_names = additional_limbs_names
        self.selected_limbs_name = selected_limbs_name
        self.logger = get_logger(logger)

        if isinstance(dataset, dict):
            dataset['logger'] = self.logger
            self.dataset = build_dataset(dataset)
        else:
            self.dataset = dataset

        if isinstance(associator, dict):
            associator['logger'] = self.logger
            self.associator = build_top_down_associator(associator)
        else:
            self.associator = associator

        if isinstance(dataset_visualization, dict):
            dataset_visualization['logger'] = self.logger
            self.dataset_visualization = build_data_visualization(
                dataset_visualization)
        else:
            self.dataset_visualization = dataset_visualization

    def run(self, overwrite: bool = False):
        prepare_output_path(
            output_path=self.output_dir,
            allowed_suffix='',
            path_type='dir',
            overwrite=overwrite,
            logger=self.logger)
        n_frame = len(self.dataset)
        n_kps = get_keypoint_num(convention=self.pred_kps3d_convention)
        pred_kps3d = np.zeros(shape=(n_frame, 1, n_kps, 4))
        matched_kps2d_idx = [[] for _ in range(n_frame)]
        gt_kps3d = None
        max_identity = 0
        end_of_clip_idxs = []
        identities = []
        for frame_idx, frame_item in enumerate(tqdm(self.dataset)):
            mview_img_tensor, _, _, _, kps3d, end_of_clip, kw_data = frame_item
            if end_of_clip:
                end_of_clip_idxs.append(frame_idx)
            fisheye_list = self.dataset.fisheye_params[0]
            mview_img_arr = np.asarray(mview_img_tensor * 255).astype(np.uint8)
            mview_kps2d_list = kw_data['kps2d']
            mview_bbox2d_list = kw_data['bbox2d']
            # prepare input for associate single frame
            mview_keypoints2d_list = []
            for view_idx, kps2d in enumerate(mview_kps2d_list):
                keypoints2d = Keypoints(
                    dtype='numpy',
                    kps=kps2d.unsqueeze(0),
                    mask=kps2d[..., -1].unsqueeze(0) > 0,
                    convention=self.dataset.kps2d_convention,
                    logger=self.logger)
                mview_keypoints2d_list.append(keypoints2d)
            self.associator.set_cameras(fisheye_list)

            keypoints2d_idx, predict_keypoints3d, identities = \
                self.associator.associate_frame(
                    mview_img_arr=mview_img_arr,
                    mview_bbox2d=mview_bbox2d_list,
                    mview_keypoints2d=mview_keypoints2d_list,
                    affinity_type='geometry_mean'
                )

            # save predict kps3d
            for idx, identity in enumerate(identities):
                if identity > max_identity:
                    n_identity = identity - max_identity
                    pred_kps3d = np.concatenate(
                        (pred_kps3d,
                         np.zeros(shape=(n_frame, n_identity, n_kps, 4))),
                        axis=1)
                    max_identity = identity
                pred_kps3d[frame_idx,
                           identity] = predict_keypoints3d.get_keypoints()[0,
                                                                           idx]
            for i in range(max_identity + 1):
                if i in identities:
                    index = identities.index(i)
                    matched_kps2d_idx[frame_idx].append(keypoints2d_idx[index])
            # save ground truth kps3d
            if gt_kps3d is None:
                gt_kps3d = kps3d.numpy()[np.newaxis]
            else:
                gt_kps3d = np.concatenate(
                    (gt_kps3d, kps3d.numpy()[np.newaxis]), axis=0)

        pred_keypoints3d = Keypoints(
            dtype='numpy',
            kps=pred_kps3d,
            mask=pred_kps3d[..., -1] > 0,
            convention=self.pred_kps3d_convention,
            logger=self.logger)
        gt_keypoints3d = Keypoints(
            dtype='numpy',
            kps=gt_kps3d,
            mask=gt_kps3d[..., -1] > 0,
            convention=self.dataset.gt_kps3d_convention,
            logger=self.logger)

        mscene_keypoints_paths = []
        scene_start_idx = 0
        for scene_idx, scene_end_idx in enumerate(end_of_clip_idxs):
            scene_keypoints = pred_keypoints3d.clone()
            kps3d = scene_keypoints.get_keypoints()[
                scene_start_idx:scene_end_idx + 1, ...]
            mask = scene_keypoints.get_mask()[scene_start_idx:scene_end_idx +
                                              1, ...]
            scene_keypoints.set_keypoints(kps3d)
            scene_keypoints.set_mask(mask)
            npz_path = osp.join(self.output_dir,
                                f'scene{scene_idx}_pred_keypoints3d.npz')
            scene_keypoints.dump(npz_path)
            scence_matched_kps2d_idx = matched_kps2d_idx[
                scene_start_idx:scene_end_idx + 1]
            np.save(
                osp.join(self.output_dir,
                         f'scene{scene_idx}_matched_kps2d_idx.npy'),
                scence_matched_kps2d_idx)
            mscene_keypoints_paths.append(npz_path)
            scene_start_idx = scene_end_idx + 1

        pred_keypoints3d_, gt_keypoints3d_, limbs = self.align_keypoints3d(
            pred_keypoints3d, gt_keypoints3d)
        self.calc_limbs_accuracy(pred_keypoints3d_, gt_keypoints3d_, limbs)
        pck_50, pck_100, mpjpe, pa_mpjpe = self.evaluate(
            pred_keypoints3d_, gt_keypoints3d_)

        if self.dataset_visualization is not None:
            self.dataset_visualization.pred_kps3d_paths = \
                mscene_keypoints_paths
            self.dataset_visualization.run(overwrite=overwrite)

    def evaluate(self,
                 pred_keypoints3d: Keypoints,
                 gt_keypoints3d: Keypoints,
                 pck_50_thres=50,
                 pck_100_thres=100,
                 scale=1000.) -> dict:
        # There must be no np.nan in the pred_keypoints3d
        mpjpe, pa_mpjpe, pck_50, pck_100 = [], [], [], []
        n_frame = gt_keypoints3d.get_frame_number()
        gt_kps3d = gt_keypoints3d.get_keypoints()[..., :3]
        gt_kps3d_mask = gt_keypoints3d.get_mask()
        pred_kps3d = pred_keypoints3d.get_keypoints()[..., :3]
        pred_kps3d_mask = pred_keypoints3d.get_mask()
        pred_kps3d_convention = pred_keypoints3d.get_convention()
        gt_kps3d_convention = gt_keypoints3d.get_convention()
        for frame_idx in range(n_frame):
            if not gt_kps3d_mask[frame_idx].any():
                continue
            gt_kps3d_idxs = np.where(
                np.sum(gt_kps3d_mask[frame_idx], axis=1) > 0)[0]
            for gt_kps3d_idx in gt_kps3d_idxs:
                f_gt_kps3d = gt_kps3d[frame_idx][gt_kps3d_idx]
                f_pred_kps3d = pred_kps3d[frame_idx][
                    np.sum(pred_kps3d_mask[frame_idx], axis=1) > 0]
                if len(f_pred_kps3d) == 0:
                    continue

                dist = vectorize_distance(f_gt_kps3d[np.newaxis], f_pred_kps3d)
                f_pred_kps3d = f_pred_kps3d[np.argmin(dist[0])]

                if np.all((f_pred_kps3d == 0)):
                    continue

                # MPJPE
                f_pred_keypoints = Keypoints(
                    kps=np.concatenate(
                        (f_pred_kps3d, np.ones_like(f_pred_kps3d[..., 0:1])),
                        axis=-1),
                    convention=pred_kps3d_convention)
                f_gt_keypoints = Keypoints(
                    kps=np.concatenate(
                        (f_gt_kps3d, np.ones_like(f_gt_kps3d[..., 0:1])),
                        axis=-1),
                    convention=gt_kps3d_convention)
                mpjpe.append(
                    compute_mpjpe(
                        f_pred_keypoints, f_gt_keypoints, align=True))

                # PA-MPJPE
                _, _, rotation, scaling, transl = compute_similarity_transform(
                    f_gt_kps3d, f_pred_kps3d, compute_optimal_scale=True)
                pred_kps3d_pa = (scaling * f_pred_kps3d.dot(rotation)) + transl

                pred_keypoints_pa = Keypoints(
                    kps=np.concatenate(
                        (pred_kps3d_pa, np.ones_like(pred_kps3d_pa[..., 0:1])),
                        axis=-1),
                    convention=pred_kps3d_convention)
                pa_mpjpe_i = compute_mpjpe(
                    pred_keypoints_pa, f_gt_keypoints, align=True)
                pa_mpjpe.append(pa_mpjpe_i)

                pck_50.append(np.mean(pa_mpjpe_i <= (pck_50_thres / scale)))
                pck_100.append(np.mean(pa_mpjpe_i <= (pck_100_thres / scale)))
        mpjpe = np.asarray(mpjpe) * scale  # m to mm
        pa_mpjpe = np.asarray(pa_mpjpe) * scale  # m to mm
        mpjpe_mean, mpjpe_std = np.mean(mpjpe), np.std(mpjpe)
        pa_mpjpe_mean, pa_mpjpe_std = np.mean(pa_mpjpe), np.std(pa_mpjpe)
        pck_50 = np.mean(pck_50) * 100.  # percentage
        pck_100 = np.mean(pck_100) * 100.  # percentage
        self.logger.info(f'MPJPE: {mpjpe_mean:.2f} ± {mpjpe_std:.2f} mm')
        self.logger.info(f'PA-MPJPE: {pa_mpjpe_mean:.2f} ±'
                         f'{pa_mpjpe_std:.2f} mm')
        self.logger.info(f'PCK@{pck_50_thres}mm: {pck_50:.2f} %')
        self.logger.info(f'PCK@{pck_100_thres}mm: {pck_100:.2f} %')
        return pck_50, pck_100, mpjpe, pa_mpjpe

    def align_keypoints3d(self, pred_keypoints3d: Keypoints,
                          gt_keypoints3d: Keypoints):
        ret_limbs = []
        gt_nose = None
        pred_nose = None
        pred_kps3d_convention = pred_keypoints3d.get_convention()
        gt_kps3d_convention = gt_keypoints3d.get_convention()
        if gt_kps3d_convention == 'panoptic':
            gt_nose_index = get_keypoint_idx(
                name='nose_openpose', convention=gt_kps3d_convention)
            gt_nose = gt_keypoints3d.get_keypoints()[:, :, gt_nose_index, :3]

        if pred_kps3d_convention == 'coco':
            pred_nose_index = get_keypoint_idx(
                name='nose', convention=pred_kps3d_convention)
            pred_nose = pred_keypoints3d.get_keypoints()[:, :,
                                                         pred_nose_index, :3]

        if pred_kps3d_convention != self.eval_kps3d_convention:
            pred_keypoints3d = convert_keypoints(
                keypoints=pred_keypoints3d,
                dst=self.eval_kps3d_convention,
                approximate=True)
        if gt_kps3d_convention != self.eval_kps3d_convention:
            gt_keypoints3d = convert_keypoints(
                keypoints=gt_keypoints3d,
                dst=self.eval_kps3d_convention,
                approximate=True)

        limbs = get_limbs_from_keypoints(
            keypoints=pred_keypoints3d, fill_limb_names=True)
        limb_name_list = []
        conn_list = []
        for limb_name, conn in limbs.get_connections_by_names().items():
            limb_name_list.append(limb_name)
            conn_list.append(conn)

        for idx, limb_name in enumerate(limb_name_list):
            if limb_name in self.selected_limbs_name:
                ret_limbs.append(conn_list[idx])

        for conn_names in self.additional_limbs_names:
            kps_idx_0 = get_keypoint_idx(
                name=conn_names[0], convention=self.eval_kps3d_convention)
            kps_idx_1 = get_keypoint_idx(
                name=conn_names[1], convention=self.eval_kps3d_convention)
            ret_limbs.append(np.array([kps_idx_0, kps_idx_1], dtype=np.int32))
        pred_kps3d_mask = pred_keypoints3d.get_mask()
        pred_kps3d = pred_keypoints3d.get_keypoints()[..., :3]
        if pred_nose is not None:
            pred_kps3d = add_campus_jaw_headtop(pred_nose, pred_kps3d)
            pred_kps3d_mask = add_campus_jaw_headtop_mask(pred_kps3d_mask)

        gt_kps3d_mask = gt_keypoints3d.get_mask()
        gt_kps3d = gt_keypoints3d.get_keypoints()[..., :3]
        if gt_nose is not None:
            gt_kps3d = add_campus_jaw_headtop(gt_nose, gt_kps3d)
            gt_kps3d_mask = add_campus_jaw_headtop_mask(gt_kps3d_mask)

        pred_kps3d = np.concatenate(
            (pred_kps3d, pred_kps3d_mask[..., np.newaxis]), axis=-1)
        pred_keypoints3d = Keypoints(
            kps=pred_kps3d,
            mask=pred_kps3d_mask,
            convention=self.eval_kps3d_convention)
        gt_kps3d = np.concatenate((gt_kps3d, gt_kps3d_mask[..., np.newaxis]),
                                  axis=-1)
        gt_keypoints3d = Keypoints(
            kps=gt_kps3d,
            mask=gt_kps3d_mask,
            convention=self.eval_kps3d_convention)

        return pred_keypoints3d, gt_keypoints3d, ret_limbs

    def calc_limbs_accuracy(self,
                            pred_keypoints3d,
                            gt_keypoints3d,
                            limbs,
                            dump_dir=None) -> Tuple[np.ndarray, list]:
        n_frame = gt_keypoints3d.get_frame_number()
        n_gt_person = gt_keypoints3d.get_person_number()
        gt_kps3d = gt_keypoints3d.get_keypoints()[..., :3]
        gt_kps3d_mask = gt_keypoints3d.get_mask()
        pred_kps3d = pred_keypoints3d.get_keypoints()[..., :3]
        pred_kps3d_mask = pred_keypoints3d.get_mask()
        check_result = np.zeros((n_frame, n_gt_person, len(limbs) + 1),
                                dtype=np.int32)
        accuracy_cnt = 0
        error_cnt = 0

        for idx in range(n_frame):
            if not gt_kps3d_mask[idx].any():
                continue
            gt_kps3d_idxs = np.where(np.sum(gt_kps3d_mask[idx], axis=1) > 0)[0]
            for gt_kps3d_idx in gt_kps3d_idxs:
                f_gt_kps3d = gt_kps3d[idx][gt_kps3d_idx]
                f_pred_kps3d = pred_kps3d[idx][
                    np.sum(pred_kps3d_mask[idx], axis=1) > 0]
                if len(f_pred_kps3d) == 0:
                    continue

                dist = vectorize_distance(f_gt_kps3d[np.newaxis], f_pred_kps3d)
                f_pred_kps3d = f_pred_kps3d[np.argmin(dist[0])]

                for i, limb in enumerate(limbs):
                    start_point, end_point = limb
                    if check_limb_is_correct(f_pred_kps3d[start_point],
                                             f_pred_kps3d[end_point],
                                             f_gt_kps3d[start_point],
                                             f_gt_kps3d[end_point]):
                        check_result[idx, gt_kps3d_idx, i] = 1
                        accuracy_cnt += 1
                    else:
                        check_result[idx, gt_kps3d_idx, i] = -1
                        error_cnt += 1
                gt_hip = (f_gt_kps3d[2] + f_gt_kps3d[3]) / 2
                pred_hip = (f_pred_kps3d[2] + f_pred_kps3d[3]) / 2
                if check_limb_is_correct(pred_hip, f_pred_kps3d[12], gt_hip,
                                         f_gt_kps3d[12]):
                    check_result[idx, gt_kps3d_idx, -1] = 1
                    accuracy_cnt += 1
                else:
                    check_result[idx, gt_kps3d_idx, -1] = -1
                    error_cnt += 1
        bone_group = dict([('Head', np.array([8])), ('Torso', np.array([9])),
                           ('Upper arms', np.array([5, 6])),
                           ('Lower arms', np.array([4, 7])),
                           ('Upper legs', np.array([1, 2])),
                           ('Lower legs', np.array([0, 3]))])

        person_wise_avg = np.sum(
            check_result > 0, axis=(0, 2)) / np.sum(
                np.abs(check_result), axis=(0, 2))

        bone_wise_result = dict()
        bone_person_wise_result = dict()
        for k, v in bone_group.items():
            bone_wise_result[k] = np.sum(check_result[:, :, v] > 0) / np.sum(
                np.abs(check_result[:, :, v]))
            bone_person_wise_result[k] = np.sum(
                check_result[:, :, v] > 0, axis=(0, 2)) / np.sum(
                    np.abs(check_result[:, :, v]), axis=(0, 2))

        tb = PrettyTable()
        tb.field_names = ['Bone Group'] + [
            f'Actor {i}'
            for i in range(bone_person_wise_result['Head'].shape[0])
        ] + ['Average']
        list_tb = [tb.field_names]
        for k, v in bone_person_wise_result.items():
            this_row = [k] + [np.char.mod('%.4f', i) for i in v
                              ] + [np.char.mod('%.4f',
                                               np.sum(v) / len(v))]
            list_tb.append([
                float(i) if isinstance(i, type(np.array([]))) else i
                for i in this_row
            ])
            tb.add_row(this_row)
        this_row = ['Total'] + [
            np.char.mod('%.4f', i) for i in person_wise_avg
        ] + [
            np.char.mod('%.4f',
                        np.sum(person_wise_avg) / len(person_wise_avg))
        ]
        tb.add_row(this_row)
        list_tb.append([
            float(i) if isinstance(i, type(np.array([]))) else i
            for i in this_row
        ])
        if dump_dir:
            np.save(
                osp.join(
                    dump_dir,
                    time.strftime('%Y_%m_%d_%H_%M',
                                  time.localtime(time.time()))), check_result)
            with open(
                    osp.join(
                        dump_dir,
                        time.strftime('%Y_%m_%d_%H_%M.csv',
                                      time.localtime(time.time()))), 'w') as f:
                writer = csv.writer(f)
                writer.writerows(list_tb)
        self.logger.info('\n' + tb.get_string())
        return check_result, list_tb



from mmcv.runner.hooks import Hook
from typing import List, TypeVar

from xrmocap.model.registrant.handler.base_handler import BaseInput

_SMPLify = TypeVar('_SMPLify')


class SMPLifyBaseHook(Hook):
    """Base class of any SMPLify hook."""

    def before_optimize(self, registrant: _SMPLify,
                        input_list: List[BaseInput],
                        optim_param: dict) -> None:
        """When SMPLify is called, this hook will be triggered just before
        optimization starts.

        Args:
            registrant (_SMPLify):
                The Registrant instance.
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            optim_param (dict):
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
                Its keys can be found in
                registrant.__class__.OPTIM_PARAM.
        """
        pass

    def after_optimize(self, registrant: _SMPLify, input_list: List[BaseInput],
                       optim_param: dict) -> None:
        """When SMPLify is called, this hook will be triggered just after
        optimization ends.

        Args:
            registrant (_SMPLify):
                The Registrant instance.
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            optim_param (dict):
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
                Its keys can be found in
                registrant.__class__.OPTIM_PARAM.
        """
        pass

    def before_stage(self, registrant: _SMPLify, input_list: List[BaseInput],
                     optim_param: dict, stage_config: dict) -> None:
        """This hook will be triggered at the beginning of
        SMPLify.__optimize_stage__().

        Args:
            registrant (_SMPLify):
                The Registrant instance.
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            optim_param (dict):
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
                Its keys can be found in
                registrant.__class__.OPTIM_PARAM.
            stage_config (dict):
                A dictionary that includes stage configurations.
        """
        pass

    def after_stage(self, registrant: _SMPLify, input_list: List[BaseInput],
                    optim_param: dict, stage_config: dict,
                    loss_dict: dict) -> None:
        """This hook will be triggered at the end of
        SMPLify.__optimize_stage__().

        Args:
            registrant (_SMPLify):
                The Registrant instance.
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            optim_param (dict):
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
                Its keys can be found in
                registrant.__class__.OPTIM_PARAM.
            stage_config (dict):
                A dictionary that includes stage configurations.
            loss_dict (dict):
                A dict that contains all losses of the last evaluation.
        """
        pass

    def before_evaluate(self, registrant: _SMPLify,
                        input_list: List[BaseInput],
                        optim_param: dict) -> None:
        """This hook will be triggered at the beginning of SMPLify.evaluate().

        Args:
            registrant (_SMPLify):
                The Registrant instance.
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            optim_param (dict):
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
                Its keys can be found in
                registrant.__class__.OPTIM_PARAM.
        """
        pass

    def after_evaluate(self, registrant: _SMPLify, input_list: List[BaseInput],
                       optim_param: dict, loss_dict: dict) -> None:
        """This hook will be triggered at the end of SMPLify.evaluate().

        Args:
            registrant (_SMPLify):
                The Registrant instance.
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            optim_param (dict):
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
                Its keys can be found in
                registrant.__class__.OPTIM_PARAM.
            loss_dict (dict):
                A dict that contains all losses of this evaluation.
        """
        pass



from setuptools import find_packages, setup


def readme():
    with open('README.md', encoding='utf-8') as f:
        content = f.read()
    return content


def get_version():
    version_file = 'xrmocap/version.py'
    with open(version_file, 'r', encoding='utf-8') as f:
        exec(compile(f.read(), version_file, 'exec'))
    return locals()['__version__']


def parse_requirements(fname='requirements.txt', with_version=True):
    """Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    Args:
        fname (str): path to requirements file
        with_version (bool, default=False): if True include version specs

    Returns:
        List[str]: list of requirements items

    CommandLine:
        python -c "import setup; print(setup.parse_requirements())"
    """
    import re
    import sys
    from os.path import exists
    require_fpath = fname

    def parse_line(line):
        """Parse information from a line in a requirements text file."""
        if line.startswith('-r '):
            # Allow specifying requirements in other files
            target = line.split(' ')[1]
            for info in parse_require_file(target):
                yield info
        else:
            info = {'line': line}
            if line.startswith('-e '):
                info['package'] = line.split('#egg=')[1]
            else:
                # Remove versioning from the package
                pat = '(' + '|'.join(['>=', '==', '>']) + ')'
                parts = re.split(pat, line, maxsplit=1)
                parts = [p.strip() for p in parts]

                info['package'] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    if ';' in rest:
                        # Handle platform specific dependencies
                        # http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-platform-specific-dependencies
                        version, platform_deps = map(str.strip,
                                                     rest.split(';'))
                        info['platform_deps'] = platform_deps
                    else:
                        version = rest  # NOQA
                    info['version'] = (op, version)
            yield info

    def parse_require_file(fpath):
        with open(fpath, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith('#'):
                    for info in parse_line(line):
                        yield info

    def gen_packages_items():
        if exists(require_fpath):
            for info in parse_require_file(require_fpath):
                parts = [info['package']]
                if with_version and 'version' in info:
                    parts.extend(info['version'])
                if not sys.version.startswith('3.4'):
                    # apparently package_deps are broken in 3.4
                    platform_deps = info.get('platform_deps')
                    if platform_deps is not None:
                        parts.append(';' + platform_deps)
                item = ''.join(parts)
                yield item

    packages = list(gen_packages_items())
    return packages


setup(
    name='xrmocap',
    version=get_version(),
    description='',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='',
    author_email='',
    keywords='',
    url='https://github.com/openxrlab/xrmocap',
    packages=find_packages(exclude=('configs', 'tools', 'demo')),
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    license='Apache License 2.0',
    tests_require=parse_requirements('requirements/test.txt'),
    install_requires=parse_requirements('requirements/build.txt'),
    zip_safe=False)



# yapf: disable
import logging
import torch
from typing import List, Union

from xrmocap.data_structure.body_model import SMPLData
from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.model.registrant.builder import SMPLify, build_registrant
from xrmocap.model.registrant.handler.builder import build_handler
from xrmocap.transform.convention.keypoints_convention import convert_keypoints
from xrmocap.transform.keypoints3d.optim.builder import (
    BaseOptimizer, build_keypoints3d_optimizer,
)
from .base_estimator import BaseEstimator

# yapf: enable


class MultiPersonSMPLEstimator(BaseEstimator):
    """Api for estimating smpl in a multi-person scene."""

    def __init__(self,
                 work_dir: str,
                 smplify: Union[dict, SMPLify],
                 kps3d_optimizers: Union[List[Union[BaseOptimizer, dict]],
                                         None] = None,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Initialization of the class.

        Args:
            work_dir (str):
                Path to the folder for running the api. No file in work_dir
                will be modified
                or added by MultiViewSinglePersonSMPLEstimator.
            smplify (Union[dict, SMPLify]):
                A SMPLify instance or its config.
            kps3d_optimizers (Union[
                    List[Union[BaseOptimizer, dict]], None], optional):
                A list of keypoints3d optimizers or their configs. If given,
                keypoints3d will be
                optimized by the cascaded final optimizers before estimation.
                Defaults to None.
            verbose (bool, optional):
                Whether to print(logger.info) information during estimating.
                Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__(work_dir, verbose, logger)

        if isinstance(smplify, dict):
            smplify['logger'] = logger
            self.smplify = build_registrant(smplify)
        else:
            self.smplify = smplify
        if kps3d_optimizers is None:
            self.kps3d_optimizers = None
        else:
            self.kps3d_optimizers = []
            for kps3d_optim in kps3d_optimizers:
                if isinstance(kps3d_optim, dict):
                    kps3d_optim['logger'] = logger
                    kps3d_optim = build_keypoints3d_optimizer(kps3d_optim)
                self.kps3d_optimizers.append(kps3d_optim)

    def optimize_keypoints3d(self, keypoints3d: Keypoints,
                             **optim_kwargs) -> Keypoints:
        """Optimize keypoints3d.

        Args:
            keypoints3d (Keypoints): A keypoints3d Keypoints instance
        Returns:
            Keypoints: The optimized keypoints3d.
        """
        if self.kps3d_optimizers is not None:
            for optimizer in self.kps3d_optimizers:
                keypoints3d = optimizer.optimize_keypoints3d(
                    keypoints3d, **optim_kwargs)
        return keypoints3d

    def estimate_smpl(self, keypoints3d: Keypoints) -> SMPLData:
        """Estimate smpl parameters according to keypoints3d.

        Args:
            keypoints3d (Keypoints):
                A keypoints3d Keypoints instance, with only one person
                inside. This method will take the person at
                keypoints3d.get_keypoints()[:, 0, ...] to run smplify.

        Returns:
            SMPLData:
                SMPL data of the person.
        """
        self.logger.info('Estimating SMPL.')
        working_convention = 'smpl'

        n_frame = keypoints3d.get_frame_number()
        n_person = keypoints3d.get_person_number()
        keypoints3d = keypoints3d.to_tensor(device=self.smplify.device)
        person_mask = keypoints3d.get_mask()
        person_mask = torch.sum(person_mask, dim=2) > 0

        keypoints3d = convert_keypoints(
            keypoints=keypoints3d, dst=working_convention)
        kps3d_tensor = keypoints3d.get_keypoints()[:, :, :, :3].float()
        kps3d_conf = keypoints3d.get_mask()[:, :, ...]

        smpl_data_list = []
        for person in range(n_person):
            if person_mask[:, person].sum() == 0:
                continue
            global_orient = torch.zeros((n_frame, 3)).to(self.smplify.device)
            transl = torch.full((n_frame, 3), 1000.0).to(self.smplify.device)
            body_pose = torch.zeros((n_frame, 69)).to(self.smplify.device)
            betas = torch.zeros((n_frame, 10)).to(self.smplify.device)
            s_kps3d_tensor = kps3d_tensor[:, person][person_mask[:, person]]
            s_kps3d_conf = kps3d_conf[:, person][person_mask[:, person]]
            # build and run
            kp3d_mse_input = build_handler(
                dict(
                    type='Keypoint3dMSEInput',
                    keypoints3d=s_kps3d_tensor,
                    keypoints3d_conf=s_kps3d_conf,
                    keypoints3d_convention=working_convention,
                    handler_key='keypoints3d_mse'))
            kp3d_llen_input = build_handler(
                dict(
                    type='Keypoint3dLimbLenInput',
                    keypoints3d=s_kps3d_tensor,
                    keypoints3d_conf=s_kps3d_conf,
                    keypoints3d_convention=working_convention,
                    handler_key='keypoints3d_limb_len'))
            registrant_output = self.smplify(
                input_list=[kp3d_mse_input, kp3d_llen_input])

            smpl_data = SMPLData()
            global_orient[
                person_mask[:, person]] = registrant_output['global_orient']
            transl[person_mask[:, person]] = registrant_output['transl']
            body_pose[person_mask[:, person]] = registrant_output['body_pose']
            betas[person_mask[:, person]] = registrant_output['betas']
            output = {}
            output['global_orient'] = global_orient
            output['transl'] = transl
            output['body_pose'] = body_pose
            output['betas'] = betas
            smpl_data.from_param_dict(output)
            smpl_data_list.append(smpl_data)
        return smpl_data_list
